# 'construct.py' contains the basic minimum spanning tree algorithms for
# constructing the tree and infering statistics.

import numpy as np
from scipy.sparse.csgraph import minimum_spanning_tree
from sklearn.neighbors import kneighbors_graph

from . import scale_cut
from . import tomo


def construct_mst(x, y, z=None, k_neighbours=20, two_dimensions=False, scale_cut_length=0., is_tomo=False):
    """Constructs the minimum spanning tree (MST) of an input data set.

    Parameters
    ----------
    x, y, z : array
        The coordinate positions of the nodes.
    k_neighbours : integer, optional
        The number of nearest neighbours to consider when creating the k-nearest neighbour graph.
    two_dimensions : bool, optional
        If true assumes the data set is two dimensional.
    scale_cut_length : float, optional
        The minimum allowed length in the k_nearest_neighbour_graph.
    is_tomo : bool, optional
        Should be true if the entered x, y and z coordinates are on a unit sphere.

    Returns
    -------
    edge_length : array
        The length of the edges included in the MST.
    edge_x, edge_y, edge_z : array
        The positional coordinates of the ends of each edge.
    edge_index : array
        The node index of the ends of each edge.
    num_removed_edges : int, optional
        Number of removed edges. Only given and applicable if scale_cut_length is given.
    """
    if two_dimensions is True:
        vertices = np.array([x, y]).T
    else:
        vertices = np.array([x, y, z]).T
    if scale_cut_length != 0.:
        if two_dimensions is True:
            x, y, k_nearest_neighbour_graph, num_removed_edges = \
                scale_cut.k_nearest_neighbour_scale_cut(x, y, scale_cut_length, k_neighbours)
        else:
            x, y, z, k_nearest_neighbour_graph, num_removed_edges = \
                scale_cut.k_nearest_neighbour_scale_cut(x, y, scale_cut_length, k_neighbours, z=z)
    else:
        k_nearest_neighbour_graph = kneighbors_graph(vertices, n_neighbors=k_neighbours, mode='distance')
    if is_tomo is True:
        k_nearest_neighbour_graph = tomo.convert_tomo_knn_length2angle(k_nearest_neighbour_graph, len(x))
    tree = minimum_spanning_tree(k_nearest_neighbour_graph, overwrite=True)
    tree = tree.tocoo()
    edge_length = tree.data
    index1 = tree.row
    index2 = tree.col
    x1 = x[index1]
    x2 = x[index2]
    edge_x = np.array([x1, x2])
    y1 = y[index1]
    y2 = y[index2]
    edge_y = np.array([y1, y2])
    if two_dimensions is False:
        z1 = z[index1]
        z2 = z[index2]
        edge_z = np.array([z1, z2])
    edge_index = np.array([index1, index2])
    if scale_cut_length == 0.:
        if two_dimensions is True:
            return edge_length, edge_x, edge_y, edge_index
        else:
            return edge_length, edge_x, edge_y, edge_z, edge_index
    else:
        if two_dimensions is True:
            return edge_length, edge_x, edge_y, edge_index, num_removed_edges
        else:
            return edge_length, edge_x, edge_y, edge_z, edge_index, num_removed_edges



# 'get_mst_class.py' contains the main class function for constructing a minimum
# spanning tree from a given set of points.

import os
import sys

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(dir_path[:-3] + '/coordinates/')

import numpy as np
import coordinate_utility
from . import construct
from . import stats
from . import branches
from . import density as dens
from . import partition


class GetMST:
    """A class for constructing and analysing the minimum spanning tree (MST). Input the
    node positions of a given data set to initiate the class.
    """

    def __init__(self, x=None, y=None, z=None, phi=None, theta=None, ra=None, dec=None,
                 r=None, units='degrees', do_print=False):
        """
        Parameters
        ----------
        x, y, (z) : array
            Cartesian 2D (3D) coordinates.
        phi, theta, (r) : array
            Tomographic (spherical) coordinates.
        ra, dec, (r) : array
            Celestial tomographic (spherical) coordinates.
        units : {'degree', 'radians'}, optional
            The units of the celestial coordinates ra and dec.
        do_print : bool, optional
            Tells the functions whether it is okay for it to print out statements.

        Notes
        -----
        The default of all of the input parameters are set to 'None' such that an internal parameter '_mode'
        of the MST can be set based on the input parameters. Supply:
            * x and y - for 2D cartesian coordinate. "_mode='2D'"
            * x, y and z - for 3D cartesian coordinates.  "_mode='3D'"
            * phi and theta - for tomographic coordinates. "_mode='tomographic'"
            * phi, theta and r - for spherical polar coordinates. "_mode='spherical polar'"
            * ra and dec - for celestial coordinates. "_mode='tomographic celestial'"
            * ra, dec and r - for celestial spherical polar coordinates. "_mode='spherical polar celestial'"
        """
        self.x = x
        self.y = y
        self.z = z
        self.phi = phi
        self.theta = theta
        self.ra = ra
        self.dec = dec
        self.r = r
        self.units = units
        self.do_print = do_print
        if self.x is not None and self.y is not None:
            if self.z is None:
                self._mode = '2D'
            else:
                self._mode = '3D'
        elif self.phi is not None and self.theta is not None:
            if self.r is None:
                self._mode = 'tomographic'
            else:
                self._mode = 'spherical polar'
        elif self.ra is not None and self.dec is not None:
            if self.r is None:
                self._mode = 'tomographic celestial'
            else:
                self._mode = 'spherical polar celestial'
        if do_print is True:
            print('MST mode: ', self._mode, ' coordinates')
        self.k_neighbours = 20
        self.edge_length = None
        self.edge_x = None
        self.edge_y = None
        self.edge_z = None
        self.edge_phi = None
        self.edge_theta = None
        self.edge_index = None
        self.degree = None
        self.mean_degree = None
        self.edge_degree = None
        self.branch_index = None
        self.branch_length = None
        self.branch_edge_count = None
        self.branch_shape = None
        self.scale_cut_length = 0.
        self.num_removed_edges_fraction = None

    def define_k_neighbours(self, k_neighbours):
        """Sets the k_neighbours value. This is automatically set to 20 if this is not called.

        Parameters
        ----------
        k_neighbours : int
            The number of nearest neighbours to consider when creating the k-nearest neighbour graph.
        """
        self.k_neighbours = k_neighbours

    def scale_cut(self, scale_cut_length):
        """Defines the scale cut parameters if a scaling cut is required.

        Parameters
        ----------
        scale_cut_length : float
            The minimum allowed length in the k_nearest_neighbour_graph.
        """
        self.scale_cut_length = scale_cut_length

    def construct_mst(self):
        """Constructs the minimum spanning tree from the input data set."""
        if self._mode == '2D':
            if self.scale_cut_length == 0.:
                edge_length, edge_x, edge_y, edge_index = \
                    construct.construct_mst(self.x, self.y, k_neighbours=self.k_neighbours, two_dimensions=True)
            else:
                edge_length, edge_x, edge_y, edge_index, num_removed_edges_fraction = \
                    construct.construct_mst(self.x, self.y, k_neighbours=self.k_neighbours, two_dimensions=True,
                                            scale_cut_length=self.scale_cut_length)
        else:
            if self._mode == 'tomographic':
                x, y, z = \
                    coordinate_utility.spherical_2_unit_sphere(self.phi, self.theta, units=self.units)
                self.x, self.y, self.z = x, y, z
            elif self._mode == 'spherical polar':
                x, y, z = \
                    coordinate_utility.spherical_2_cartesian(self.r, self.phi, self.theta, units=self.units)
                self.x, self.y, self.z = x, y, z
            elif self._mode == 'tomographic celestial':
                phi, theta, x, y, z = \
                    coordinate_utility.celestial_2_unit_sphere(self.ra, self.dec, units=self.units, output='both')
                self.x, self.y, self.z = x, y, z
                self.phi, self.theta = phi, theta
            elif self._mode == 'spherical polar celestial':
                phi, theta, x, y, z = \
                    coordinate_utility.celestial_2_cartesian(self.r, self.ra, self.dec, units=self.units, output='both')
                self.x, self.y, self.z = x, y, z
                self.phi, self.theta = phi, theta
            else:
                pass
            if self.scale_cut_length == 0.:
                edge_length, edge_x, edge_y, edge_z, edge_index = \
                    construct.construct_mst(self.x, self.y, z=self.z, k_neighbours=self.k_neighbours,
                                            two_dimensions=False)
            else:
                edge_length, edge_x, edge_y, edge_z, edge_index, num_removed_edges_fraction = \
                    construct.construct_mst(self.x, self.y, z=self.z, k_neighbours=self.k_neighbours,
                                            two_dimensions=False, scale_cut_length=self.scale_cut_length)
            self.edge_z = edge_z
        self.edge_length, self.edge_x, self.edge_y, self.edge_index = edge_length, edge_x, edge_y, edge_index
        if self.phi is not None:
            self.edge_phi = np.array([self.phi[self.edge_index[0]], self.phi[self.edge_index[1]]])
        if self.theta is not None:
            self.edge_theta = np.array([self.theta[self.edge_index[0]], self.theta[self.edge_index[1]]])
        if self.scale_cut_length != 0.:
            self.num_removed_edges_fraction = num_removed_edges_fraction

    def get_degree(self):
        """Finds the degree of each node in the constructed MST."""
        if self.edge_index is not None:
            self.degree = stats.get_graph_degree(self.edge_index, len(self.x))
        else:
            raise ValueError("'edge_index' are undefined, meaning the minimum spanning tree has yet to be constructed.")

    def get_mean_degree_for_edges(self):
        """Finds the mean degree for each edge.
        """
        if self.degree is not None:
            self.mean_degree = stats.get_mean_degree_for_edges(self.edge_index, self.degree)
        else:
            raise ValueError("The degrees are undefined, meaning they have yet to be calculated.")

    def get_degree_for_edges(self):
        """Gets the degree of the nodes at each end of all edge."""
        if self.degree is not None:
            self.edge_degree = stats.get_degree_for_edges(self.edge_index, self.degree)
        else:
            raise ValueError("The degrees are undefined, meaning they have yet to be calculated.")

    def get_branches(self, box_size=None, sub_divisions=1):
        """Finds the branches of a MST.

        Parameters
        ----------
        box_size : float, optional
            The size of the '2D' or '3D' box.
        sub_divisions : int, optional
            The number of divisions used to divide the data set in each axis. Used for speeding up the branch
            finding algorithm when using many points (> 100000).
        """
        if sub_divisions is 1:
            branch_index, rejected_branch_index = branches.get_branch_index(self.edge_index, self.edge_degree)
        else:
            if self._mode is '2D':
                branch_index, rejected_branch_index = \
                    branches.get_branch_index_sub_divide(sub_divisions, self.edge_index, self.edge_degree,
                                                         box_size=box_size, edge_x=self.edge_x, edge_y=self.edge_y,
                                                         mode='Euclidean', two_dimension=True)
            elif self._mode is '3D':
                branch_index, rejected_branch_index = \
                    branches.get_branch_index_sub_divide(sub_divisions, self.edge_index, self.edge_degree,
                                                         box_size=box_size, edge_x=self.edge_x, edge_y=self.edge_y,
                                                         edge_z=self.edge_z, mode='Euclidean', two_dimension=False)
            else:
                branch_index, rejected_branch_index = \
                    branches.get_branch_index_sub_divide(sub_divisions, self.edge_index, self.edge_degree,
                                                         box_size=None, phi=self.phi, theta=self.theta,
                                                         edge_phi=self.edge_phi, edge_theta=self.edge_theta, mode='spherical')
        self.branch_index = branch_index
        if len(rejected_branch_index) is not 0:
            if self.do_print is True:
                print(str(float(len(rejected_branch_index))) + ' branches were incompleted.')
        branch_length = [np.sum(self.edge_length[i]) for i in branch_index]
        self.branch_length = np.array(branch_length)

    def get_branch_edge_count(self):
        """Finds the number of edges included in each branch."""
        branch_edge_count = [float(len(i)) for i in self.branch_index]
        self.branch_edge_count = np.array(branch_edge_count)

    def get_branch_shape(self):
        """Finds the shape of all branches. This is simply the straight line distance between the two ends divided by
        the branch length."""
        self.branch_shape = branches.get_branch_shape(self.edge_index, self.edge_degree, self.branch_index,
                                                      self.branch_length, mode=self._mode, x=self.x, y=self.y, z=self.z)

    def get_stats_vs_density(self, dx, box_size):
        """Computes the relation between the density contrast and the MST statistics.

        Parameters
        ----------
        dx : float
            The length of the individual cells, that the full box will be divided into, across one dimension.
        box_size : float
            The length of the 2D or 3D box across one axis.

        Returns
        -------
        density : array
            Density contrast of each cell.
        mean_degree, mean_edge_length, mean_branch_length, mean_branch_shape : array
            The mean of the 'degree', 'edge length', 'branch length' and 'branch shape' in each respective cell.

        To do
        -----
        Add support for data sets given in 'tomographic' and 'spherical polar' coordinates.
        """
        if self._mode == '2D':
            mean_degree, density = dens.variable_vs_density(self.x, self.y, dx, self.x, self.y, self.degree, box_size,
                                                            mode='2D', get_density=True)
            x_edge_mean = 0.5 * (self.edge_x[0] + self.edge_x[1])
            y_edge_mean = 0.5 * (self.edge_y[0] + self.edge_y[1])
            mean_edge_length = dens.variable_vs_density(self.x, self.y, dx, x_edge_mean, y_edge_mean, self.edge_length,
                                                        box_size, mode='2D')
            branch_index_end = branches.get_branch_end_index(self.edge_index, self.edge_degree, self.branch_index)
            x_branch_mean = 0.5 * (self.x[branch_index_end[0]] + self.x[branch_index_end[1]])
            y_branch_mean = 0.5 * (self.y[branch_index_end[0]] + self.y[branch_index_end[1]])
            mean_branch_length = dens.variable_vs_density(self.x, self.y, dx, x_branch_mean, y_branch_mean,
                                                          self.branch_length, box_size, mode='2D')
            mean_branch_shape = dens.variable_vs_density(self.x, self.y, dx, x_branch_mean, y_branch_mean,
                                                         self.branch_shape, box_size, mode='2D')
            return density, mean_degree, mean_edge_length, mean_branch_length, mean_branch_shape
        elif self._mode == '3D':
            mean_degree, density = dens.variable_vs_density(self.x, self.y, dx, self.x, self.y, self.degree, box_size,
                                                            z=self.z, z_param=self.z, mode='3D', get_density=True)
            x_edge_mean = 0.5 * (self.edge_x[0] + self.edge_x[1])
            y_edge_mean = 0.5 * (self.edge_y[0] + self.edge_y[1])
            z_edge_mean = 0.5 * (self.edge_z[0] + self.edge_z[1])
            mean_edge_length = dens.variable_vs_density(self.x, self.y, dx, x_edge_mean, y_edge_mean, self.edge_length,
                                                        box_size, z=self.z, z_param=z_edge_mean, mode='3D')
            branch_index_end = branches.get_branch_end_index(self.edge_index, self.edge_degree, self.branch_index)
            x_branch_mean = 0.5 * (self.x[branch_index_end[0]] + self.x[branch_index_end[1]])
            y_branch_mean = 0.5 * (self.y[branch_index_end[0]] + self.y[branch_index_end[1]])
            z_branch_mean = 0.5 * (self.z[branch_index_end[0]] + self.z[branch_index_end[1]])
            mean_branch_length = dens.variable_vs_density(self.x, self.y, dx, x_branch_mean, y_branch_mean,
                                                          self.branch_length, box_size, z=self.z, z_param=z_branch_mean,
                                                          mode='3D')
            mean_branch_shape = dens.variable_vs_density(self.x, self.y, dx, x_branch_mean, y_branch_mean,
                                                         self.branch_shape, box_size, z=self.z, z_param=z_branch_mean,
                                                         mode='3D')
            return density, mean_degree, mean_edge_length, mean_branch_length, mean_branch_shape
        else:
            raise ValueError("Computation for 'Spherical polar' or 'tomographic' data sets is currently unsupported.")

    def output_stats(self, include_index=False):
        """Outputs the MST statistics.

        Parameters
        ----------
        include_index : bool, optional
            If true will output the indexes of the nodes for each edge and the indexes of edges in each branch.

        Returns
        -------
        degree : array
            The degree of each node in the MST.
        edge_length : array
            The length of each edge in the MST.
        branch_length : array
            The length of branches in the MST.
        branch_shape : array
            The shape of branches in the MST.
        edge_index : array, optional
            A 2 dimensional array, where the first nested array shows the indexes for the nodes
            on one end of the edge and the second shows the other node.
        branch_index : list, optional
            A list of branches, where each branch is given as a list of the indexes of the member edges.
        """
        if include_index is True:
            return self.degree, self.edge_length, self.branch_length, self.branch_shape, self.edge_index, \
                   self.branch_index
        else:
            return self.degree, self.edge_length, self.branch_length, self.branch_shape

    def _get_stats(self, include_index=False, sub_divisions=1, k_neighbours=None, scale_cut_length=0.):
        """Computes the MST and outputs the statistics.

        Parameters
        ----------
        include_index : bool, optional
            If True will output the indexes of the nodes for each edge and the indexes of edges in each branch.
        sub_divisions : int, optional
            The number of divisions used to divide the data set in each axis. Used for speeding up the branch
            finding algorithm when using many points (> 100000).
        k_neighbours : int, optional
            The number of nearest neighbours to consider when creating the k-nearest neighbour graph.
        scale_cut_length : float, optional
            The minimum allowed length in the k_nearest_neighbour_graph.

        Returns
        -------
        degree : array
            The degree of each node in the MST.
        edge_length : array
            The length of each edge in the MST.
        branch_length : array
            The length of branches in the MST.
        branch_shape : array
            The shape of branches in the MST.
        edge_index : array, optional
            A 2 dimensional array, where the first nested array shows the indexes for the nodes
            on one end of the edge and the second shows the other node.
        branch_index : list, optional
            A list of branches, where each branch is given as a list of the indexes of the member edges.

        Notes
        -----
        This will calculate all the MST statistics by putting the data set through the following functions:
            1) k_neighbours (if k_neighbours is specified)
            2) construct_mst
            3) get_degree
            4) get_degree_for_edges
            5) get_branches
            6) get_branch_shape
            7) output_stats
        """
        if k_neighbours is not None:
            self.define_k_neighbours(k_neighbours)
        if scale_cut_length != 0.:
            self.scale_cut(scale_cut_length=scale_cut_length)
        self.construct_mst()
        self.get_degree()
        self.get_degree_for_edges()
        self.get_branches(sub_divisions=sub_divisions)
        self.get_branch_shape()
        return self.output_stats(include_index=include_index)

    def get_stats(self, include_index=False, sub_divisions=1, k_neighbours=None,
                  scale_cut_length=0., partitions=1):
        """Gets the minimum spanning tree statistics of a partitioned data set. Same inputs as 'get_stats'.

        Parameters
        ----------
        sub_divisions : int, optional
            The number of divisions used to divide the data set in each axis.
        k_neighbours : int, optional
            The number of nearest neighbours to consider when creating the k-nearest neighbour graph.
        scale_cut_length : float, optional
            The minimum allowed length in the k_nearest_neighbour_graph.
        partitions : int
            Number of partitions to divide the data set into.

        Returns
        -------
        degree : array
            The degree of each node in the MST.
        edge_length : array
            The length of each edge in the MST.
        branch_length : array
            The length of branches in the MST.
        branch_shape : array
            The shape of branches in the MST.
        edge_index : list, optional
            A list of 2 dimensional arrays for the nodes in each group.
        branch_index : list, optional
            A list of list of branches, where each branch is given as a list of the indexes of the member edges.
        groups : array, optional
            The assigned groups for each point in the data set (only outputed if include_index=True). Indexes
            here are indexes of the elements in each group.
        """
        if float(partitions) == 1.:
            return self._get_stats(include_index=include_index, sub_divisions=sub_divisions,
                                   k_neighbours=k_neighbours, scale_cut_length=scale_cut_length)
        else:
            # check mode and store all points in temporary array with prefix 'all_'.
            if self._mode == '2D':
                all_x = np.copy(self.x)
                all_y = np.copy(self.y)
                total_sample_size = len(all_x)
            elif self._mode == '3D':
                all_x = np.copy(self.x)
                all_y = np.copy(self.y)
                all_z = np.copy(self.z)
                total_sample_size = len(all_x)
            elif self._mode == 'tomographic':
                all_phi = np.copy(self.phi)
                all_theta = np.copy(self.theta)
                total_sample_size = len(all_phi)
            elif self._mode == 'spherical polar':
                all_phi = np.copy(self.phi)
                all_theta = np.copy(self.theta)
                all_r = np.copy(self.r)
                total_sample_size = len(all_phi)
            elif self._mode == 'tomographic celestial':
                all_ra = np.copy(self.ra)
                all_dec = np.copy(self.dec)
                total_sample_size = len(all_ra)
            elif self._mode == 'spherical polar celestial':
                all_ra = np.copy(self.ra)
                all_dec = np.copy(self.dec)
                all_r = np.copy(self.r)
                total_sample_size = len(all_ra)
            else:
                pass
            # get groupings
            groups = partition.partition_data(total_sample_size, partitions)
            edge_index = []
            branch_index = []
            for i in range(0, partitions):
                group_indices = partition.get_index_for_group(groups, i)
                if self._mode == '2D':
                    self.__init__(x=all_x[group_indices], y=all_y[group_indices])
                elif self._mode == '3D':
                    self.__init__(x=all_x[group_indices], y=all_y[group_indices], z=all_z[group_indices])
                elif self._mode == 'tomographic':
                    self.__init__(phi=all_phi[group_indices], theta=all_theta[group_indices])
                elif self._mode == 'spherical polar':
                    self.__init__(phi=all_phi[group_indices], theta=all_theta[group_indices], r=all_r[group_indices])
                elif self._mode == 'tomographic celestial':
                    self.__init__(ra=all_ra[group_indices], dec=all_dec[group_indices])
                elif self._mode == 'spherical polar celestial':
                    self.__init__(ra=all_ra[group_indices], dec=all_dec[group_indices], r=all_r[group_indices])
                else:
                    pass
                if include_index is False:
                    _degree, _edge_length , _branch_length, _branch_shape = \
                        self._get_stats(sub_divisions=sub_divisions, k_neighbours=k_neighbours, scale_cut_length=scale_cut_length)
                else:
                    _degree, _edge_length , _branch_length, _branch_shape, _edge_index, _branch_index = \
                        self._get_stats(include_index=True, sub_divisions=sub_divisions, k_neighbours=k_neighbours, scale_cut_length=scale_cut_length)
                if i == 0:
                    degree = _degree
                    edge_length = _edge_length
                    branch_length = _branch_length
                    branch_shape = _branch_shape
                else:
                    degree = np.concatenate([degree, _degree])
                    edge_length = np.concatenate([edge_length, _edge_length])
                    branch_length = np.concatenate([branch_length, _branch_length])
                    branch_shape = np.concatenate([branch_shape, _branch_shape])
                if include_index is True:
                    edge_index.append(_edge_index)
                    branch_index.append(_branch_index)
            if include_index is True:
                return degree, edge_length, branch_length, branch_shape, edge_index, branch_index, groups
            else:
                return degree, edge_length, branch_length, branch_shape
            if self._mode == '2D':
                self.x = all_x
                self.y = all_y
            elif self._mode == '3D':
                self.x = all_x
                self.y = all_y
                self.z = all_z
            elif self._mode == 'tomographic':
                self.phi = all_phi
                self.theta = all_theta
            elif self._mode == 'spherical polar':
                self.phi = all_phi
                self.theta = all_theta
                self.r = all_r
            elif self._mode == 'tomographic celestial':
                self.ra = all_ra
                self.dec = all_dec
            elif self._mode == 'spherical polar celestial':
                self.ra = all_ra
                self.dec = all_dec
                self.r = all_r
            else:
                pass

    def clean(self):
        self.x = None
        self.y = None
        self.z = None
        self.ra = None
        self.dec = None
        self.r = None
        self.units = None
        self.do_print = False
        self._mode = None
        self.k_neighbours = 20
        self.phi = None
        self.theta = None
        self.edge_length = None
        self.edge_x = None
        self.edge_y = None
        self.edge_z = None
        self.edge_phi = None
        self.edge_theta = None
        self.edge_index = None
        self.degree = None
        self.mean_degree = None
        self.edge_degree = None
        self.branch_index = None
        self.branch_length = None
        self.branch_edge_count = None
        self.branch_shape = None
        self.scale_cut_length = 0.
        self.num_removed_edges_fraction = None



name = "mistree"

# coordinate utility functions
from .coordinates.coordinate_utility import spherical_2_cartesian
from .coordinates.coordinate_utility import celestial_2_cartesian
from .coordinates.coordinate_utility import spherical_2_unit_sphere
from .coordinates.coordinate_utility import celestial_2_unit_sphere
from .coordinates.coordinate_utility import perpendicular_distance_2_angle

# levy flight distributions
from .levy_flight.levy_flight import get_random_flight
from .levy_flight.levy_flight import get_levy_flight
from .levy_flight.levy_flight import get_adjusted_levy_flight

# mst - branch functions
from .mst.branches import get_branch_index
from .mst.branches import get_branch_index_sub_divide
from .mst.branches import get_branch_end_index
from .mst.branches import get_branch_edge_count
from .mst.branches import get_branch_shape

# mst - construction functions
from .mst.construct import construct_mst

# mst - density vs variable functions
from .mst.density import variable_vs_density

# mst - graph functions
from .mst.graph import graph2data
from .mst.graph import data2graph

# mst - histogram functions
from .mst.hist_mst import bin_data
from .mst.hist_mst import HistMST

# mst - plotting functions
from .mst.plot_mst import set_plot_default
from .mst.plot_mst import plot_histogram_line
from .mst.plot_mst import plot_histogram_confidence
from .mst.plot_mst import plot_histogram_error
from .mst.plot_mst import PlotHistMST

# mst - scale cut functions
from .mst.scale_cut import graph_scale_cut
from .mst.scale_cut import graph_scale_cut
from .mst.scale_cut import k_nearest_neighbour_scale_cut

# mst - single class functions
from .mst.get_mst_class import GetMST

# mst - mst statistical functions
from .mst.stats import get_graph_degree
from .mst.stats import get_mean_degree_for_edges
from .mst.stats import get_degree_for_edges

# mst - tomographic functions
from .mst.tomo import convert_tomo_knn_length2angle



import numpy as np
import matplotlib.pylab as plt


def bin_data(data, minimum=None, maximum=None, bin_size=None, bin_number=100, normalised=True):
    """Returns the (normalised) number count of a data set with values within defined bins.

    Parameters
    ----------
    data : array_like
        The data to be binned.
    minimum : float
        A given minimum for the bin edges.
    maximum : float
        A given maximum for the bin edges
    bin_size : float
        The size of the bins.
    bin_number : int
        The number of bins to be used.
    normalised : bool
        Tf true will return a normalised histogram, if false returns a number counts.

    Returns
    -------
    bin_centres : array_like
        The central value of each bin.
    binned_data : array_like
        The binned number count (or normalised histogram) of the data set.
    """
    if minimum is None:
        minimum = np.min(data)
    if maximum is None:
        maximum = np.max(data)
    if bin_size is None:
        _bin_edge = np.linspace(minimum, maximum, bin_number+1)
    if bin_size is not None:
        _bin_edge = np.arange(minimum, maximum+bin_size, bin_size)
        condition = np.where(_bin_edge <= maximum)[0]
        _bin_edge = _bin_edge[condition]
    binned_data, _bin_edges = np.histogram(data, bins=_bin_edge, density=normalised)
    bin_centres = 0.5 * (_bin_edge[1:] + _bin_edge[:-1])
    return bin_centres, binned_data


class HistMST:

    def __init__(self):
        """Initialises MST histogram class.
        """
        # bin edges information
        self.uselog = None
        self.use_s_sqrt = None
        self.usenorm = None
        self.d_min = None
        self.d_max = None
        self.num_d_bins = None
        self.l_min = None
        self.l_max = None
        self.num_l_bins = None
        self.b_min = None
        self.b_max = None
        self.num_b_bins = None
        self.s_min = None
        self.s_max = None
        self.num_s_bins = None
        self.logl_min = None
        self.logl_max = None
        self.logb_min = None
        self.logb_max = None
        self.group_mode = False

    def setup(self, uselog=False, use_sqrt_s=True, usenorm=True, d_min=0.5, d_max=6.5, num_d_bins=6,
              l_min=0., l_max=None, num_l_bins=100, b_min=0., b_max=None, num_b_bins=100,
              s_min=0., s_max=1., num_s_bins=50, logl_min=None, logl_max=None, logb_min=None, logb_max=None):
        """Setups bin sizes for the MST statistics.

        Parameters
        ----------
        uselog : bool
            Determines whether to use log bins for l and b.
        use_sqrt_s : bool
            Determines whether to use the sqrt(1-s) projection of s or just s itself.
        usenorm : bool
            Determines whether to normalise the histograms.
        d_min : float
            Minimum for degree bins (use half integer values).
        d_max : float
            Maximum for degree bins (use half integer values).
        num_d_bins : int
            Number of bins for the distribution of degree, this should be equal to d_max - d_min.
        l_min : float
            Minimum for edge length bins.
        l_max : float
            Maximum for edge length bins.
        num_l_bins : int
            Number of bins for the distribution of edge lengths.
        b_min : float
            Minimum for branch length bins.
        b_max : float
            Maximum for branch length bins.
        num_b_bins : int
            Number of bins for the distribution of branch lengths.
        s_min : float
            Minimum for branch shape bins.
        s_max : float
            Maximum for branch shape bins.
        num_s_bins : int
            Number of bins for the distribution of branch shapes.
        logl_min : float
            Minimum of edge lengths in log to base 10.
        logl_max : float
            Maximum of edge lengths in log to base 10.
        logb_min : float
            Minimum of branch lengths in log to base 10.
        logb_max : float
            Maximum of branch lengths in log to base 10.
        """
        self.uselog = uselog
        self.use_sqrt_s= use_sqrt_s
        self.usenorm = usenorm
        self.d_min = d_min
        self.d_max = d_max
        self.num_d_bins = num_d_bins
        self.l_min = l_min
        self.l_max = l_max
        self.num_l_bins = num_l_bins
        self.b_min = b_min
        self.b_max = b_max
        self.num_b_bins = num_b_bins
        self.s_min = s_min
        self.s_max = s_max
        self.num_s_bins = num_s_bins
        self.logl_min = logl_min
        self.logl_max = logl_max
        self.logb_min = logb_min
        self.logb_max = logb_max

    def get_hist(self, d, l, b, s):
        """Bins the MST distribution which is returned as a dictionary.

        Parameters
        ----------
        d : array_like
            Distribution of degree.
        l : array_like
            Distribution of edge length.
        b : array_like
            Distribution of branch length.
        s : array_like
            Distribution of branch shape.

        Returns
        -------
        mst_hist : dict
            Dictionary of MST binned histograms.
        """
        # find minimum and maximum
        if self.l_max is None:
            self.l_max = 1.05*l.max()
        if self.b_max is None:
            self.b_max = 1.05*b.max()
        if self.logl_min is None:
            self.logl_min = np.log10(0.95*l.min())
        if self.logl_max is None:
            self.logl_max = np.log10(1.05*l.max())
        if self.logb_min is None:
            self.logb_min = np.log10(0.95*b.min())
        if self.logb_max is None:
            self.logb_max = np.log10(1.05*b.max())
        # bin mst statistics
        x_d, y_d = bin_data(d, minimum=self.d_min, maximum=self.d_max, bin_number=self.num_d_bins, normalised=self.usenorm)
        if self.uselog == False:
            x_l, y_l = bin_data(l, minimum=self.l_min, maximum=self.l_max, bin_number=self.num_l_bins, normalised=self.usenorm)
            x_b, y_b = bin_data(b, minimum=self.b_min, maximum=self.b_max, bin_number=self.num_b_bins, normalised=self.usenorm)
        else:
            x_logl, y_l = bin_data(np.log10(l), minimum=self.logl_min, maximum=self.logl_max, bin_number=self.num_l_bins, normalised=self.usenorm)
            x_logb, y_b = bin_data(np.log10(b), minimum=self.logb_min, maximum=self.logb_max, bin_number=self.num_b_bins, normalised=self.usenorm)
            x_l = 10.**x_logl
            x_b = 10.**x_logb
        if self.use_sqrt_s == False:
            x_s, y_s = bin_data(s, minimum=self.s_min, maximum=self.s_max, bin_number=self.num_s_bins, normalised=self.usenorm)
        else:
            x_s, y_s = bin_data(np.sqrt(1.-s), minimum=self.s_min, maximum=self.s_max, bin_number=self.num_s_bins, normalised=self.usenorm)
        mst_hist = {
            "uselog" : self.uselog, "use_sqrt_s" : self.use_sqrt_s, "usenorm" : self.usenorm, "isgroup" : False,
            "x_d" : x_d, "y_d" : y_d, "x_l" : x_l, "y_l" : y_l, "x_b" : x_b, "y_b" : y_b, "x_s" : x_s, "y_s" : y_s
        }
        if self.group_mode == True:
            self.x_d = mst_hist['x_d']
            self.x_l = mst_hist['x_l']
            self.x_b = mst_hist['x_b']
            self.x_s = mst_hist['x_s']
            self.group_y_d.append(mst_hist['y_d'])
            self.group_y_l.append(mst_hist['y_l'])
            self.group_y_b.append(mst_hist['y_b'])
            self.group_y_s.append(mst_hist['y_s'])
        return mst_hist

    def start_group(self):
        """Begins group mode for calculating the mean and standard deviation of the MST
        from different realisations of data points coming from the same model.
        """
        self.group_mode = True
        self.group_y_d = []
        self.group_y_l = []
        self.group_y_b = []
        self.group_y_s = []

    def end_group(self):
        """Ends group mode for calculating the mean and standard deviation of the MST.

        Returns
        -------
        mst_hist : dict
            Dictionary of the mean and standard deviation of the MST binned histograms.
        """
        self.group_mode = False
        self.group_y_d = np.array(self.group_y_d)
        self.group_y_l = np.array(self.group_y_l)
        self.group_y_b = np.array(self.group_y_b)
        self.group_y_s = np.array(self.group_y_s)
        y_d_mean = np.mean(self.group_y_d, axis=0)
        y_l_mean = np.mean(self.group_y_l, axis=0)
        y_b_mean = np.mean(self.group_y_b, axis=0)
        y_s_mean = np.mean(self.group_y_s, axis=0)
        y_d_std = np.std(self.group_y_d, axis=0)
        y_l_std = np.std(self.group_y_l, axis=0)
        y_b_std = np.std(self.group_y_b, axis=0)
        y_s_std = np.std(self.group_y_s, axis=0)
        mst_hist = {
            "uselog" : self.uselog, "use_sqrt_s" : self.use_sqrt_s,
            "usenorm" : self.usenorm, "isgroup" : True,
            "x_d" : self.x_d, "y_d" : y_d_mean, "y_d_std" : y_d_std,
            "x_l" : self.x_l, "y_l" : y_l_mean, "y_l_std" : y_l_std,
            "x_b" : self.x_b, "y_b" : y_b_mean, "y_b_std" : y_b_std,
            "x_s" : self.x_s, "y_s" : y_s_mean, "y_s_std" : y_s_std,
        }
        return mst_hist

    def clean(self):
        """Resets HistMST variables.
        """
        self.__init__()



# 'tomo.py' converts tomographic k-nearest neighbour graphs from being defined
# across a unitary sphere to perpendical angles across a sphere.

import os
import sys

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(dir_path[:-3] + '/coordinates/')

from . import graph
import coordinate_utility


def convert_tomo_knn_length2angle(k_nearest_neighbour_graph, number_of_nodes):
    """Converts the k-nearest neighbour graph from 3D lengths on a unit sphere to angular distances.

    Parameters
    ----------
    k_nearest_neighbour_graph : csr_matrix
        A sparse matrix of the nearest k neighbours.
    number_of_nodes : int
        The number of nodes from which the nearest neighbour graph is constructed.

    Returns
    -------
    k_nearest_neighbour_graph_angle : csr_matrix
        The same as input with edges returned in angles (radians).
    """
    index1, index2, distances = graph.graph2data(k_nearest_neighbour_graph)
    distances_angles = coordinate_utility.perpendicular_distance_2_angle(distances)
    k_nearest_neighbour_graph_angle = graph.data2graph(index1, index2, distances_angles, number_of_nodes)
    return k_nearest_neighbour_graph_angle



# 'branches.py' finds and measures the length and shapes of branches in from a
# given constructed MST.

import os
import sys

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(dir_path[:-3] + '/coordinates/')

import numpy as np
import coordinate_utility


def get_branch_index(edge_index, edge_degree, branch_cutting_frequency=1000):
    """Finds the branch indexes for each branch in the MST.

    Parameters
    ----------
    edge_index : array
        The node index of the ends of each edge.
    edge_degree : array
        The degree for the ends of each edge.
    branch_cutting_frequency : int, optional
        An optimisation parameter, used to remove edges that have already been placed into a branch.
        This significantly improves the speed of the algorithm as branches that are already constructed
        are now removed from the branch finder.

    Returns
    -------
    branch_index : list
        A list of branches where each branch is a list of the edge index of edges contained in each branch.
    branch_index_rejected : list
        A list of branches that have not been completed. This will occur only if a subset of the edge indexes
        of the full tree is provided.
    """
    degree1 = edge_degree[0]
    degree2 = edge_degree[1]
    index1 = edge_index[0]
    index2 = edge_index[1]
    condition = np.where((degree1 == 2.) & (degree2 == 2.))[0]
    index_branch_mid = condition
    index_branch_mid1 = index1[index_branch_mid]
    index_branch_mid2 = index2[index_branch_mid]
    condition = np.where(((degree1 == 2.) & (degree2 != 2.)) | ((degree1 != 2.) & (degree2 == 2.)))[0]
    index_branch_end = condition
    index_branch_end1 = index1[index_branch_end]
    index_branch_end2 = index2[index_branch_end]
    degree_branch_end1 = degree1[index_branch_end]
    degree_branch_end2 = degree2[index_branch_end]
    check_mid = np.ones(len(index_branch_mid))
    check_end = np.ones(len(index_branch_end))
    branch_index = []
    branch_index_rejected = []
    mask_end = np.ones(index_branch_end.shape, dtype=np.bool)
    mask_mid = np.ones(index_branch_mid.shape, dtype=np.bool)
    count = 0
    item = 0
    while item < len(index_branch_end):
        if check_end[item] == 1.:
            check_end[item] = 0.
            done = 0.
            _twig = []
            _twig.append(index_branch_end[item])
            if degree_branch_end1[item] == 2.:
                node_index = index_branch_end1[item]
            elif degree_branch_end2[item] == 2.:
                node_index = index_branch_end2[item]
            else:
                assert ValueError("branch edge incorrect.")
            mask_end[item] = False
            while done == 0.:
                condition = np.where(((check_mid == 1.) & (index_branch_mid1 == node_index)) |
                                     ((check_mid == 1.) & (index_branch_mid2 == node_index)))[0]
                if len(condition) == 0:
                    condition = np.where(((check_end == 1.) & (index_branch_end1 == node_index)) |
                                         ((check_end == 1.) & (index_branch_end2 == node_index)))[0]
                    if len(condition) == 0:
                        branch_index_rejected = branch_index_rejected + \
                                                np.ndarray.tolist(np.ndarray.flatten(np.array(_twig)))
                        done = 1.
                    else:
                        check_end[condition] = 0.
                        _twig.append(index_branch_end[condition])
                        done = 1.
                        mask_end[condition] = False
                        branch_index.append(np.ndarray.tolist(np.ndarray.flatten(np.array(_twig))))
                else:
                    if len(condition) == 1:
                        check_mid[condition] = 0.
                        _twig.append(index_branch_mid[condition])
                        if index_branch_mid1[condition] == node_index:
                            node_index = index_branch_mid2[condition]
                        elif index_branch_mid2[condition] == node_index:
                            node_index = index_branch_mid1[condition]
                        else:
                            assert ValueError("Identification error.")
                        mask_mid[condition] = False
                    else:
                        assert ValueError("Found more than one vertex.")
        else:
            pass
        if count % branch_cutting_frequency == 0 and count != 0:
            index_branch_end = index_branch_end[mask_end]
            check_end = check_end[mask_end]
            index_branch_end1 = index_branch_end1[mask_end]
            index_branch_end2 = index_branch_end2[mask_end]
            degree_branch_end1 = degree_branch_end1[mask_end]
            degree_branch_end2 = degree_branch_end2[mask_end]
            index_branch_mid = index_branch_mid[mask_mid]
            check_mid = check_mid[mask_mid]
            index_branch_mid1 = index_branch_mid1[mask_mid]
            index_branch_mid2 = index_branch_mid2[mask_mid]
            mask_end = mask_end[mask_end]
            mask_mid = mask_mid[mask_mid]
            count = count + 1
            item = 0
        elif count % 1001 == 0:
            count = count + 1
            item = item + 1
        elif item == len(index_branch_end) - 1:
            index_branch_end = index_branch_end[mask_end]
            check_end = check_end[mask_end]
            index_branch_end1 = index_branch_end1[mask_end]
            index_branch_end2 = index_branch_end2[mask_end]
            degree_branch_end1 = degree_branch_end1[mask_end]
            degree_branch_end2 = degree_branch_end2[mask_end]
            index_branch_mid = index_branch_mid[mask_mid]
            check_mid = check_mid[mask_mid]
            index_branch_mid1 = index_branch_mid1[mask_mid]
            index_branch_mid2 = index_branch_mid2[mask_mid]
            mask_end = mask_end[mask_end]
            mask_mid = mask_mid[mask_mid]
            count = count + 1
            item = 0
        else:
            count = count + 1
            item = item + 1
    branch_index_rejected = branch_index_rejected + np.ndarray.tolist(np.ndarray.flatten(np.array(index_branch_mid)))
    branch_index = [np.ndarray.tolist(np.hstack(np.array(branch_index[i]))) for i in range(0, len(branch_index))]
    if len(branch_index_rejected) != 0:
        branch_index_rejected = np.ndarray.tolist(np.hstack(np.array(branch_index_rejected)))
    return branch_index, branch_index_rejected


def get_branch_index_sub_divide(sub_divisions, edge_index, edge_degree, box_size=None, edge_x=None, edge_y=None,
                                edge_z=None, phi=None, theta=None, edge_phi=None, edge_theta=None,
                                branch_cutting_frequency=1000, mode='Euclidean', two_dimension=True):
    """ Finds the length of branches for large sets of data where a rapid increase in speed is achieved by subdividing
    the full data set and finding branches in each sub division and then completing branches that straddle across the
    sub divides.

    Parameters
    ----------
    sub_divisions : int
        The number of divisions used to divide the data set in each axis. A significant boost in speed is achieved.
    edge_degree : array
        The degree for the ends of each edge.
    edge_index : array
        The node index of the ends of each edge.
    box_size : {'None', float}, optional
        The size of the '2D' or '3D' box. If undefined, this will be determined by the maximum x value.
    edge_x, edge_y, edge_z : array
        The cartesian coordinates of the nodes at each end of every edge.
    phi, theta : array
        The spherical coordinates of the nodes on the sphere.
    edge_phi, edge_theta : array
        The spherical coordinates of the nodes at each end of every edge.
    branch_cutting_frequency : int
        An optimisation parameter, used to remove edges that have already been placed into a branch.
        This significantly improves the speed of the algorithm as branches that are already constructed
        are now removed from the branch finder.
    mode : {'2D', '3D', string}, optional
        '2D', '3D' or assumed to be in spherical coordinates.
    two_dimension : bool, optional
        Determines whether the data set is 2D of 3D.

    Returns
    -------
    branch_index : list
        A list of branches where each branch is a list of the edge index of edges contained in each branch.
    branch_index_rejected : list
        A list of branches that have not been completed. This will occur only if a subset of the edge indexes
        of the full tree is provided.
    """
    if mode == 'Euclidean':
        x_min, x_max, y_min, y_max = np.min(edge_x), np.max(edge_x), np.min(edge_y), np.max(edge_y)
        dx, dy = (x_max - x_min) / float(sub_divisions), (y_max - y_min) / float(sub_divisions)
        xx = np.arange(x_min, x_max + dx, dx)
        yy = np.arange(y_min, y_max + dy, dy)
        xx_mid = 0.5 * (xx[1:len(xx)] + xx[0:len(xx) - 1])
        yy_mid = 0.5 * (yy[1:len(yy)] + yy[0:len(yy) - 1])
        if two_dimension is True:
            x_div, y_div = np.meshgrid(xx_mid, yy_mid, indexing='ij')
            x_div = np.ndarray.flatten(x_div)
            y_div = np.ndarray.flatten(y_div)
        else:
            z_min, z_max = np.min(edge_z), np.max(edge_z)
            dz = (z_max - z_min) / float(sub_divisions)
            zz = np.arange(z_min, z_max + dz, dz)
            zz_mid = 0.5 * (zz[1:len(zz)] + zz[0:len(zz) - 1])
            x_div, y_div, z_div = np.meshgrid(xx_mid, yy_mid, zz_mid, indexing='ij')
            x_div = np.ndarray.flatten(x_div)
            y_div = np.ndarray.flatten(y_div)
            z_div = np.ndarray.flatten(z_div)
    else:
        phi_min, phi_max, theta_min, theta_max = np.min(phi), np.max(phi), np.min(theta), np.max(theta)
        dphi, dtheta = (phi_max - phi_min) / float(sub_divisions), (theta_max - theta_min) / float(sub_divisions)
        phiphi = np.arange(phi_min, phi_max + dphi, dphi)
        thetatheta = np.arange(theta_min, theta_max + dtheta, dtheta)
        pp = 0.5 * (phiphi[1:len(phiphi)] + phiphi[0:len(phiphi) - 1])
        tt = 0.5 * (thetatheta[1:len(thetatheta)] + thetatheta[0:len(thetatheta) - 1])
        phi_div, theta_div = np.meshgrid(pp, tt, indexing='ij')
        phi_div = np.ndarray.flatten(phi_div)
        theta_div = np.ndarray.flatten(theta_div)
    branch_index_total = []
    branch_index_rejected_total = []
    if mode == 'Euclidean':
        length = len(x_div)
        total_mask = np.ones(len(edge_x[0]))
    else:
        length = len(phi_div)
        total_mask = np.ones(len(edge_phi[0]))
    for k in range(0, length):
        if mode == 'Euclidean' and two_dimension is True:
            xd, yd = x_div[k], y_div[k]
            condition = np.where((total_mask == 1.) & ((edge_x[0] >= xd - dx / 2.) | (edge_x[0] <= xd + dx / 2.) |
                                 (edge_y[0] >= yd - dx / 2.) | (edge_y[0] <= yd + dx / 2.)))[0]
        elif mode == 'Euclidean' and two_dimension is False:
            xd, yd, zd = x_div[k], y_div[k], z_div[k]
            condition = np.where((total_mask == 1.) & ((edge_x[0] >= xd - dx / 2.) | (edge_x[0] <= xd + dx / 2.) |
                                 (edge_y[0] >= yd - dx / 2.) | (edge_y[0] < yd + dx / 2.) | (edge_z[0] >= zd - dx / 2.)
                                 & (edge_z[0] <= zd + dx / 2.)))[0]
        else:
            pd, td = phi_div[k], theta_div[k]
            condition = np.where((total_mask == 1.) & ((edge_phi[0] >= pd - dphi / 2.) | (edge_phi[0] <= pd + dphi / 2.)
                                 | (edge_theta[0] >= td - dtheta / 2.) | (edge_theta[0] <= td + dtheta / 2.)))[0]
        edge_degree_cut = np.array([edge_degree[0][condition], edge_degree[1][condition]])
        edge_index_cut = np.array([edge_index[0][condition], edge_index[1][condition]])
        branch_index_cut, branch_index_rejected_cut = \
            get_branch_index(edge_index_cut, edge_degree_cut, branch_cutting_frequency=branch_cutting_frequency)
        branch_index_cut_corrected = [np.ndarray.tolist(condition[i]) for i in branch_index_cut]
        branch_index_total = branch_index_total + branch_index_cut_corrected
        total_mask[[item for sublist in branch_index_total for item in sublist]] = 0.
        if len(branch_index_rejected_cut) is not 0:
            branch_index_rejected_total = branch_index_rejected_total + \
                                          np.ndarray.tolist(condition[branch_index_rejected_cut])
        total_mask[branch_index_rejected_total] = 0.
    branch_index_rejected_total = np.array(branch_index_rejected_total)
    branch_index_rejected_total = np.unique(branch_index_rejected_total)
    if len(branch_index_rejected_total) == 0:
        pass
    else:
        edge_degree_rejected = np.array([edge_degree[0][branch_index_rejected_total],
                                         edge_degree[1][branch_index_rejected_total]])
        edge_index_rejected = np.array([edge_index[0][branch_index_rejected_total],
                                        edge_index[1][branch_index_rejected_total]])
        branch_index_left_over, branch_index_rejected_left_over = \
            get_branch_index(edge_index_rejected, edge_degree_rejected, branch_cutting_frequency)
        branch_index_left_over_corrected = \
            [np.ndarray.tolist(branch_index_rejected_total[i]) for i in branch_index_left_over]
        branch_index_total = branch_index_total + branch_index_left_over_corrected
        branch_index_rejected_total = branch_index_rejected_left_over
    return branch_index_total, branch_index_rejected_total


def get_branch_end_index(edge_index, edge_degree, branch_index):
    """Gets the index of the nodes at the extreme end of each branch.

    Parameters
    ----------
    edge_index : array
        The node index of the ends of each edge.
    edge_degree : array
        The degree for the ends of each edge.
    branch_index : list
        A list of branches. Listing the indices of edges within each branch.

    Returns
    -------
    branch_end : array
        The index of the nodes at the ends of each branch.
    """
    branch_edge_index_end1 = [i[0] for i in branch_index]
    branch_edge_index_end2 = [i[len(i) - 1] for i in branch_index]
    edge_degree_end12 = edge_degree[1][branch_edge_index_end1]
    index11 = edge_index[0][branch_edge_index_end1]
    index12 = edge_index[1][branch_edge_index_end1]
    condition = np.where(edge_degree_end12 != 2.)[0]
    branch_index_end1 = np.copy(index11)
    branch_index_end1[condition] = index12[condition]
    edge_degree22 = edge_degree[1][branch_edge_index_end2]
    index21 = edge_index[0][branch_edge_index_end2]
    index22 = edge_index[1][branch_edge_index_end2]
    condition = np.where(edge_degree22 != 2.)[0]
    branch_index_end2 = np.copy(index21)
    branch_index_end2[condition] = index22[condition]
    branch_end = np.array([branch_index_end1, branch_index_end2])
    return branch_end


def get_branch_edge_count(branch_index):
    """Finds the number of edges included in each branch.
    """
    branch_edge_count = [float(len(i)) for i in branch_index]
    branch_edge_count = np.array(branch_edge_count)
    return branch_edge_count


def get_branch_shape(edge_index, edge_degree, branch_index, branch_length, mode='2D', x=None, y=None, z=None):
    """Finds the shape of all branches. This is simply the straight line distance between the two ends divided by
    the branch length.
    """
    branch_index_end = get_branch_end_index(edge_index, edge_degree, branch_index)
    branch_index_end1, branch_index_end2 = branch_index_end[0], branch_index_end[1]
    if mode == '2D':
        dx = abs(x[branch_index_end1] - x[branch_index_end2])
        dy = abs(y[branch_index_end1] - y[branch_index_end2])
        branch_end_length = np.sqrt((dx ** 2.) + (dy ** 2.))
    elif mode == '3D' or mode == 'spherical polar':
        dx = abs(x[branch_index_end1] - x[branch_index_end2])
        dy = abs(y[branch_index_end1] - y[branch_index_end2])
        dz = abs(z[branch_index_end1] - z[branch_index_end2])
        branch_end_length = np.sqrt((dx ** 2.) + (dy ** 2.) + (dz ** 2.))
    else:
        dx = abs(x[branch_index_end1] - x[branch_index_end2])
        dy = abs(y[branch_index_end1] - y[branch_index_end2])
        dz = abs(z[branch_index_end1] - z[branch_index_end2])
        branch_end_length = np.sqrt((dx ** 2.) + (dy ** 2.) + (dz ** 2.))
        branch_end_length = coordinate_utility.perpendicular_distance_2_angle(branch_end_length)
    branch_shape = branch_end_length/branch_length
    return branch_shape



# 'levy_flight.py' has a number of functions that can generate random levy
# flight-like distributions.

import numpy as np
from . import utility_random_walk as random_walk


def _get_randoms_sphere(size):
    """Generate random points on a sphere.

    Parameters
    ----------
    size : int
        Size of random sample.

    Returns
    -------
    phi : array
        Longitude coordinate.
    theta :array
        Latitude coordinate.
    """
    if size == 1:
        phi = 2.*np.pi*np.random.random_sample(1)[0]
        u = np.random.random_sample(1)[0]
        theta = np.arccos(1.-2.*u)
    else:
        phi = 2.*np.pi*np.random.random_sample(size)
        u = np.random.random_sample(size)
        theta = np.arccos(1.-2.*u)
    return phi, theta


def get_random_flight(steps, mode='3D', box_size=75., periodic=True):
    """Generates a random realisation of a 'Levy flight'-like distribution. The random step sizes are defined by the
    user.

    Parameters
    ----------
    steps : array
        Distribution of step sizes defined by the user.
    mode : {'2D', '3D'}, optional
        Defines whether the distribution is defined in 2D or 3D cartesian coordinates.
    box_size : float
        Length of the periodic box across one axis.
    periodic : bool
        If True then this sets periodic boundary conditions on the Levy flight realisation. If False, the box_size
        parameter is ignored.

    Returns
    -------
    x, y, z : array
        Distribution of random walk particles. z is only outputed if this is a 3D distribution.

    Internal
    --------
    _x_start, _y_start, _z_start : array
        Starting position of the distribution.
    _size : int
        The size of the distribution, 1 + size of the step_sizes array.
    """
    _size = len(steps)
    if periodic is False:
        box_size = None
    if mode == '2D':
        if box_size is None:
            _x_start, _y_start = 0., 0.
            _periodic = 0
            box_size = 0.
        else:
            _x_start = np.random.uniform(0., box_size, 1)[0]
            _y_start = np.random.uniform(0., box_size, 1)[0]
            _periodic = 1
        _phi = np.random.uniform(0., 2. * np.pi, _size)
        x, y = random_walk.random_walk_fast_2d(steps, _phi, box_size, _x_start, _y_start, _periodic, _size)
        return x, y
    elif mode == '3D':
        if box_size is None:
            _x_start, _y_start, _z_start = 0., 0., 0.
            _periodic = 0
            box_size = 0.
        else:
            _x_start = np.random.uniform(0., box_size, 1)[0]
            _y_start = np.random.uniform(0., box_size, 1)[0]
            _z_start = np.random.uniform(0., box_size, 1)[0]
            _periodic = 1
        _phi, _theta = _get_randoms_sphere(_size)
        x, y, z = random_walk.random_walk_fast_3d(steps, _phi, _theta, box_size, _x_start, _y_start, _z_start,
                                                  _periodic, _size)
        return x, y, z


def get_levy_flight(size, mode='3D', periodic=True, box_size=75., t_0=0.2, alpha=1.5):
    """Generates a random realisation of a Levy flight distribution.

    Parameters
    ----------
    size : int
        Number of points.
    mode : {'2D', '3D'}, optional
        Defines whether the distribution is defined in 2D or 3D cartesian coordinates.
    box_size : float, optional
        Length of the periodic box across one axis.
    periodic : bool, optional
        If True then this sets periodic boundary condition on the Levy flight realisation. If False, the box_size parameter is ignored.
    t_0, alpha : float, optional
        Are parameters of the Levy flight model.

    Returns
    -------
    x, y, z : array
        Distribution of Levy flight particles. z is only outputted if this is a 3D distribution.

    Internal
    --------
    _u : array
        Uniform random numbers between 0 and 1.
    _steps : array
        Random step sizes.
    _x_start, _y_start, _z_start : float
        Starting position of the distribution.
    """
    _u = np.random.uniform(0., 1., size - 1)
    _steps = t_0 / (1. - _u) ** (1. / alpha)
    if periodic is False:
        box_size = None
    if mode == '2D':
        x, y = get_random_flight(_steps, mode='2D', box_size=box_size)
        return x, y
    elif mode == '3D':
        x, y, z = get_random_flight(_steps, mode='3D', box_size=box_size)
        return x, y, z
    else:
        print("Error: Mode is unsupported. Only allowed modes are '2D' or '3D'.")


def get_adjusted_levy_flight(size, mode='3D', periodic=True, box_size=75.,
                             t_0=0.325, t_s=0.015, alpha=1.5, beta=0.45, gamma=1.3):
    """Generates a random realisation of an adjusted Levy flight (ALF) distribution.

    Parameters
    ----------
    size : int
        Number of points.
    mode : {'2D', '3D'}, optional
        Defines whether the distribution is defined in 2D or 3D cartesian coordinates.
    box_size : float, optional
        Length of the periodic box across one axis.
    periodic : bool, optional
        If True then this sets periodic boundary condition on the Levy flight realisation. If False, the box_size parameter is ignored.
    t_0, t_s, alpha, beta, gamma : float
        Are parameters of the adjusted levy walk model.

    Returns
    -------
    x, y, z : array
        Distribution of adjusted Levy flight particles. z is only outputted if this is a 3D distribution.

    Internal
    --------
    _u : array
        Uniform random numbers between 0 and 1.
    _steps : array
        Random step sizes.
    _x_start, _y_start, _z_start : float
        Starting position of the distribution.
    """
    if gamma is None:
        # If gamma is not given, then it can be calculated by imposing the condition that the probability distribution
        # function remains smooth across t_0.
        gamma = alpha*((1.-beta)/beta)*((t_0-t_s)/t_0)
    else:
        pass
    _u = np.random.uniform(0., 1., size - 1)
    _steps = (t_0 - t_s) * (_u / beta) ** (1. / gamma) + t_s
    condition = np.where(_u >= beta)[0]
    _steps[condition] = t_0*(1.+((beta-_u[condition])/(1.-beta)))**(-1./alpha)
    if periodic is False:
        box_size = None
    if mode == '2D':
        x, y = get_random_flight(_steps, mode='2D', box_size=box_size)
        return x, y
    elif mode == '3D':
        x, y, z = get_random_flight(_steps, mode='3D', box_size=box_size)
        return x, y, z
    else:
        print("Error: Mode is unsupported. Only allowed modes are '2D' or '3D'.")



# 'density.py' compares the density of particles in cells to another variable.

import numpy as np
from . import utility_density


def variable_vs_density(x, y, dx, x_param, y_param, param, box_size, z=None, z_param=None, mode='2D',
                        get_density=False):
    """ Finds the relationship between a variable and the density of points.

    Parameters
    ----------
    x, y, z : array
        The positions of points.
    x_param, y_param, z_param : array
        The positions associated with the measured parameter.
    param : array
        An array of the input parameter.
    dx : float
        The size of each cell in each axis.
    box_size : float
        The size of the box along one axis.
    mode : {'2D', '3D'}
        Specified whether the data is two or three dimensional
    get_density : bool, optional
        If true will output the density of particles.

    Returns
    -------
    mean_param : array
        The mean of the parameter in each cell.
    density : array
        The density of points in each cell.

    To do
    -----
    Extend for use on tomographic or spherical polar data sets.
    """
    xx = np.arange(dx / 2., box_size + dx / 2., dx)
    condition = np.where(xx < box_size)[0]
    xx = xx[condition]
    length = len(condition)
    if mode == '2D':
        x_div, y_div = np.meshgrid(xx, xx, indexing='ij')
        x_div, y_div = np.ndarray.flatten(x_div), np.ndarray.flatten(y_div)
        mean_param = utility_density.get_param_2d(x_param, y_param, param, dx, length, len(x_div), len(x_param))
        if get_density is True:
            counts = utility_density.get_counts_2d(x, y, dx, length, len(x_div), len(x))
            density = counts / np.mean(counts)
            return mean_param, density
        else:
            return mean_param
    else:
        x_div, y_div, z_div = np.meshgrid(xx, xx, xx, indexing='ij')
        x_div, y_div, z_div = np.ndarray.flatten(x_div), np.ndarray.flatten(y_div), np.ndarray.flatten(z_div)
        mean_param = utility_density.get_param_3d(x_param, y_param, z_param, param, dx, length, len(x_div),
                                                  len(x_param))
        if get_density is True:
            counts = utility_density.get_counts_3d(x, y, z, dx, length, len(x_div), len(x))
            density = counts / np.mean(counts)
            return mean_param, density
        else:
            return mean_param



#'partition.py' is used to partition input data sets.

import numpy as np


def partition_data(total_sample_size, num_groups):
    """Returns an array with values that correspond to the assigned groupings.

    Parameters
    ----------
    total_sample_size : int
        Length of the original data set.
    num_groups: int
        The number of groups/partitions the data is divided into.

    Returns
    -------
    groups : array
        An array of the same size as the total sample, indicating which group
        each element is assigned to. Unassigned points are given a group of -1,
        which occurs if the number of groups does not exactly divide the total_sample_size
    """
    individual_sample_size = total_sample_size/num_groups
    indices = np.arange(0, total_sample_size, 1)
    groups = np.zeros(len(indices))
    for i in range(1, num_groups):
        groups[int(individual_sample_size*i):int(individual_sample_size*(i+1))] = i
    if total_sample_size % num_groups != 0:
        groups[individual_sample_size*num_groups:] = -1
    randoms = np.random.random_sample(len(groups))
    ordered_randoms_and_reordered_groups = np.array(sorted(zip(randoms, groups)))
    groups = ordered_randoms_and_reordered_groups[:, 1]
    return groups


def get_index_for_group(groups, which_group):
    """Returns the corresponding indices for a specified group/partition.

    Paramaters
    ----------
    groups : array
        An array of the assigned groupings for each element of a given data set.
    which_group : int
        The specific group we need the indices for.

    Returns
    -------
    group_indexes : array
        The indexes for a specified group.
    """
    group_indexes = np.where(groups == float(which_group))[0]
    return group_indexes



from .coordinate_utility import spherical_2_cartesian
from .coordinate_utility import celestial_2_cartesian
from .coordinate_utility import spherical_2_unit_sphere
from .coordinate_utility import celestial_2_unit_sphere
from .coordinate_utility import perpendicular_distance_2_angle



# branch functions
from .branches import get_branch_index
from .branches import get_branch_index_sub_divide
from .branches import get_branch_end_index
from .branches import get_branch_edge_count
from .branches import get_branch_shape

# construction functions
from .construct import construct_mst

# density vs variable functions
from .density import variable_vs_density

# graph functions
from .graph import graph2data
from .graph import data2graph

# Histogram MST functions & class
from .hist_mst import bin_data
from .hist_mst import HistMST

# Plotting MST class
from .plot_mst import set_plot_default
from .plot_mst import plot_histogram_line
from .plot_mst import plot_histogram_confidence
from .plot_mst import plot_histogram_error
from .plot_mst import PlotHistMST

# scale cut functions
from .scale_cut import graph_scale_cut
from .scale_cut import graph_scale_cut
from .scale_cut import k_nearest_neighbour_scale_cut

# single class functions
from .get_mst_class import GetMST

# mst statistical functions
from .stats import get_graph_degree
from .stats import get_mean_degree_for_edges
from .stats import get_degree_for_edges

# tomographic functions
from .tomo import convert_tomo_knn_length2angle



# 'stats.py' contains functions for determining statistical properties (degree
# quantities at present) of an input graph.

import numpy as np
from . import utility_mst


def get_graph_degree(edge_index, number_of_nodes):
    """Finds the degree of each node in a constructed graph.

    Parameters
    ----------
    edge_index : array
        A 2 dimensional array, containing the node indexes on either side of each edge.
    number_of_nodes : int
        The number of nodes in the graph.

    Returns
    -------
    degree : array
        The degree for each node in the graph.
    """
    index1, index2 = edge_index[0], edge_index[1]
    _number_of_edges = len(index1)
    degree = utility_mst.get_degree_for_index(index1, index2, number_of_nodes, _number_of_edges)
    return degree


def get_mean_degree_for_edges(edge_index, degree):
    """Finds the mean degree for each edge. This is done by finding the mean of the degree at each end of the edges.

    Parameters
    ----------
    edge_index : array
        A 2 dimensional array, containing the node indexes on either side of each edge.
    degree : array
        The degree for each node in the graph.

    Returns
    -------
    mean_degree : array
        The mean degree for each edge.
    """
    index1, index2 = edge_index[0], edge_index[1]
    mean_degree = 0.5 * (degree[index1] + degree[index2])
    return mean_degree


def get_degree_for_edges(edge_index, degree):
    """Gets the degree of the nodes at each end of the edge.

    Parameters
    ----------
    edge_index : array
        The node index of the ends of each edge.
    degree : array
        The degree for each node in the graph.

    Returns
    -------
    edge_degree : array
        The degree for each node in an edge.
    """
    index1, index2 = edge_index[0], edge_index[1]
    edge_degree = np.array([degree[index1], degree[index2]])
    return edge_degree



# 'coordinate_utility.py' contains a number of coordinate transformation functions.

import numpy as np


def spherical_2_cartesian(r, phi, theta, units='degrees'):
    """Converts spherical polar coordinates into cartesian coordinates.

    Parameters
    ----------
    r : array
        Radial distance.
    phi : array
        Longitudinal coordinates (radians = [0, 2*pi]).
    theta : array
        Latitude coordinates (radians = [0, pi]).
    units : str
        Units of phi and theta given in either degrees or radians.

    Returns
    -------
    x, y, z : array
        'euclidean': euclidean coordinates.
    """
    phi = np.copy(phi)
    theta = np.copy(theta)
    if units == 'degrees':
        phi, theta = np.deg2rad(phi), np.deg2rad(theta)
    elif units == 'radians':
        pass
    else:
        raise AssertionError("Unexpected value entered for 'units', only supports either degrees or radians", units)
    x = r * np.cos(phi) * np.sin(theta)
    y = r * np.sin(phi) * np.sin(theta)
    z = r * np.cos(theta)
    return x, y, z


def celestial_2_cartesian(r, ra, dec, units='degrees', output='both'):
    """Converts spherical polar coordinates into cartesian coordinates.

    Parameters
    ----------
    r : array
        Radial distance.
    ra : array
        Longitudinal celestial coordinates.
    dec : array
        Latitude celestial coordinates.
    units : str
        Units of ra and dec given in either degrees or radians.
    output : {'cartesian', 'both'}, optional
        Determines whether to output only the cartesian or both cartesian and spherical coordinates.

    Returns
    -------
    phi, theta : array
        'spherical': spherical polar coordinates.
    x, y, z : array
        'cartesian': euclidean coordinates.
    """
    phi = np.copy(ra)
    theta = np.copy(dec)
    if units == 'degrees':
        phi, theta = np.deg2rad(phi), np.deg2rad(theta)
    elif units == 'radians':
        pass
    else:
        raise AssertionError("Unexpected value entered for 'units', only supports either degrees or radians", units)
    theta = np.pi / 2. - theta
    if output == 'spherical':
        return phi, theta
    else:
        x = r * np.cos(phi) * np.sin(theta)
        y = r * np.sin(phi) * np.sin(theta)
        z = r * np.cos(theta)
        if output == 'cartesian':
            return x, y, z
        elif output == 'both':
            return phi, theta, x, y, z
        else:
            raise AssertionError("Unexpected value entered for 'output', should be either 'cartesian' or 'both'.",
                                 output)


def spherical_2_unit_sphere(phi, theta, units='degrees'):
    """Project coordinates on a sphere into cartesian coordinates on a unit sphere.

    Parameters
    ----------
    phi : array
        Longitudinal coordinates (radians = [0, 2*pi]).
    theta : array
        Latitude coordinates (radians = [0, pi]).
    units : {'degrees', 'radians'}, optional
        Units of phi and theta given in either 'degrees' or 'radians'.

    Returns
    -------
    x, y, z : array
        cartesian coordinates.
    """
    if np.isscalar(phi) is True:
        return spherical_2_cartesian(1., phi, theta, units=units)
    else:
        return spherical_2_cartesian(np.ones(len(phi)), phi, theta, units=units)


def celestial_2_unit_sphere(ra, dec, units='degrees', output='both'):
    """Project coordinates on a sphere into cartesian coordinates on a unit sphere.

    Parameters
    ----------
    ra : array
        Longitudinal celestial coordinates.
    dec : array
        Latitude celestial coordinates.
    units : {'degrees', 'radians'}, optional
        Units of ra and dec given in either 'degrees' or 'radians'.
    output : {'cartesian', 'both'}, optional
        Determines whether to output only the euclidean or both euclidean and spherical coordinates.

    Returns
    -------
    phi, theta : array
        'spherical': spherical polar coordinates.
    x, y, z : array
        'cartesian': cartesian coordinates.
    """
    if np.isscalar(ra) is True:
        return celestial_2_cartesian(1., ra, dec, units=units, output=output)
    else:
        return celestial_2_cartesian(np.ones(len(ra)), ra, dec, units=units, output=output)


def perpendicular_distance_2_angle(distance):
    """Converts distances on a unit sphere to angular distances projected across a unit sphere.

    Parameters
    ----------
    distance : array
        Perpendicular distances across (i.e. going on the surface) of a unit sphere.

    Returns
    -------
    angular_distance : array
        The angular distance of points across a unit sphere.
    """
    angular_distance = 2. * np.arcsin(distance / 2.)
    return angular_distance



"""setup.py compiles the fortran files so they can be called from python."""

import setuptools
from numpy.distutils.core import setup, Extension
"""
def readme(short=False):
    with open('README.md') as f:
        if short:
            return f.readlines()[1].strip()
        else:
            return f.read()"""

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

ext1 = Extension(name = 'mistree.levy_flight.utility_random_walk',
                 sources = ['mistree/levy_flight/utility_random_walk.f90'])
ext2 = Extension(name = 'mistree.mst.utility_density',
                 sources = ['mistree/mst/utility_density.f90'])
ext3 = Extension(name = 'mistree.mst.utility_mst',
                 sources = ['mistree/mst/utility_mst.f90'])

setup(name = 'mistree',
      version = '1.1.3',
      description       = "A python package for constructing and analysing the minimum spanning tree",
      long_description  = long_description,
      long_description_content_type = 'text/markdown',
      url               = 'https://knaidoo29.github.io/mistreedoc/',
      author            = "Krishna Naidoo",
      author_email      = "krishna.naidoo.11@ucl.ac.uk",
      license='MIT',
      packages=setuptools.find_packages(),
      install_requires=['numpy', 'matplotlib', 'scipy', 'scikit-learn'],
      ext_modules = [ext1, ext2, ext3],
      python_requires = '>=2.7',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Fortran',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      )



# 'scale_cut.py' applies a scale cut to an input graph, removing edges below a
# given scale cut limit.

import numpy as np
from scipy.sparse import csr_matrix
from sklearn.neighbors import kneighbors_graph
from . import graph as gr


def graph_scale_cut(graph, scale_cut_length, num_nodes):
    """Will remove all edges in the graph below the scale_cut_length.

    Parameters
    ----------
    graph : csr_matrix
        A sparse matrix of the edges in a graph and corresponding node indexes.
    scale_cut_length : float
        A minimum length scale.
    num_nodes : int
        Number of nodes.

    Returns
    -------
    graph_cut : csr_matrix
        The original graph with distances below the scale_cut_length removed.
    index1, index2 : array
        The node indexes of each end of the edges in the graph.
    num_removed_edges : int
        Number of removed edges.
    """
    index1, index2, distances = gr.graph2data(graph)
    condition = np.where((distances >= scale_cut_length))[0]
    num_removed_edges_fraction = float(len(index1) - len(condition))/float(len(index1))
    index1, index2, distances = index1[condition], index2[condition], distances[condition]
    graph_cut = gr.data2graph(index1, index2, distances, num_nodes)
    return graph_cut, index1, index2, num_removed_edges_fraction


def k_nearest_neighbour_scale_cut(x, y, scale_cut_length, k_neighbours, z=None):
    """Iteratively removes edges below the scale_cut_length of a k_nearest_neighbour graph.

    Parameters
    ----------
    x, y, (z) : array
        2D or 3D coordinates of the positions of the nodes.
    scale_cut_length : float
        The minimum allowed length in the k_nearest_neighbour_graph.
    k_neighbours : int
        The number of nearest neighbours to consider when creating the k-nearest neighbour graph.
    Returns
    -------
    x, y, (z) : array
        The 2D or 3D coordinates of the positions of the nodes.
    knn : csr_matrix
        A sparse scale cut k_nearest_neighbour_graph.
    """
    if z is None:
        vertices = np.array([x, y]).T
    else:
        vertices = np.array([x, y, z]).T
    knn = kneighbors_graph(vertices, n_neighbors=k_neighbours, mode='distance')
    knn, index1, index2, num_removed_edges_fraction = graph_scale_cut(knn, scale_cut_length, len(x))
    if z is None:
        return x, y, knn, num_removed_edges_fraction
    else:
        return x, y, z, knn, num_removed_edges_fraction



import numpy as np
import matplotlib.pylab as plt

from . import hist_mst

def set_plot_default(use_bold=True):
    """Sets the default fonts for the matplotlib plots.

    Parameters
    ----------
    use_bold : bool
        Determines whether the fonts used are in bold.
    """
    plt.rc('text', usetex=True)
    plt.rc('font', family='serif')
    if use_bold == True:
        plt.rcParams['text.latex.preamble'] = [r'\boldmath']


def plot_histogram_line(x_mid, histogram, x_edges=None, ax=None, color='C0',
                        alpha=1., linewidth=1., linestyle='-'):
    """Outputs the histogram line boxes without using the histogram function.

    Parameters
    ----------
    x_mid : array_like
        Midpoint of each histogram bar.
    histogram : array_like
        Histogram height.
    x_edges : array_like
        The edges of each histogram bar. Ideal for unequal spacings between histograms.
    ax : class
        The matplotlib plotting axis.
    color : str
        The color of the histogram error plot.
    alpha : float
        The transparency fraction of the histogram error plot.
    linewidth : float
        The linewidth of histogram line.
    linestyle : str
        The linestyle used.
    """
    x_hist = []
    y_hist = []
    if x_edges is None:
        _dx = x_mid[1]-x_mid[0]
        for i in range(0,len(histogram)):
            x_hist.append(x_mid[i] - 0.5*_dx)
            x_hist.append(x_mid[i] + 0.5*_dx)
            y_hist.append(histogram[i])
            y_hist.append(histogram[i])
    else:
        for i in range(0, len(histogram)):
            x_hist.append(x_edges[i])
            x_hist.append(x_edges[i+1])
            y_hist.append(histogram[i])
            y_hist.append(histogram[i])
    x_hist, y_hist = np.array(x_hist), np.array(y_hist)
    if ax == None:
        plt.plot(x_hist, y_hist, color=color, alpha=alpha,
                 linewidth=linewidth, linestyle=linestyle)
    else:
        ax.plot(x_hist, y_hist, color=color, alpha=alpha,
                linewidth=linewidth, linestyle=linestyle)


def plot_histogram_confidence(x_mid, histogram_min, histogram_max, x_edges=None,
                              ax=None, color='C0', alpha=0.25):
    """Plots the confidence envelopes for a histogram.

    Parameters
    ----------
    x_mid : array_like
        Midpoint of each histogram bar.
    histogram_min : array_like
        Histogram minimum envelope.
    histogram_max : array_like
        Histogram maximum envelope.
    x_edges : array_like
        The edges of each histogram bar. Ideal for unequal spacings between histograms.
    ax : class
        The matplotlib plotting axis.
    color : str
        The color of the histogram error plot.
    alpha : float
        The transparency fraction of the histogram error plot.
    """
    if x_edges is None:
        _dx = x_mid[1]-x_mid[0]
        for i in range(0, len(histogram_min)):
            if ax == None:
                plt.fill_between(np.array([x_mid[i]-0.5*_dx, x_mid[i]+0.5*_dx]),
                                np.array([histogram_min[i], histogram_min[i]]),
                                np.array([histogram_max[i], histogram_max[i]]),
                                color=color,alpha=alpha)
            else:
                ax.fill_between(np.array([x_mid[i]-0.5*_dx, x_mid[i]+0.5*_dx]),
                                np.array([histogram_min[i], histogram_min[i]]),
                                np.array([histogram_max[i], histogram_max[i]]),
                                color=color,alpha=alpha)
    else:
        for i in range(0, len(histogram_min)):
            if ax == None:
                plt.fill_between(np.array([x_edges[i], x_edges[i+1]]),
                                np.array([histogram_min[i], histogram_min[i]]),
                                np.array([histogram_max[i], histogram_max[i]]),
                                color=color,alpha=alpha)
            else:
                ax.fill_between(np.array([x_edges[i], x_edges[i+1]]),
                                np.array([histogram_min[i], histogram_min[i]]),
                                np.array([histogram_max[i], histogram_max[i]]),
                                color=color,alpha=alpha)


def plot_histogram_error(x_mid, histogram, histogram_error, x_edges=None, ax=None,
                         color='C0', alpha=0.25):
    """Plots the errorbars envelopes for a histogram.

    Parameters
    ----------
    x_mid : array_like
        Midpoint of each histogram bar.
    histogram : array_like
        Histogram.
    histogram_error : array_like
        The error associated to each histogram bar.
    x_edges : array_like
        The edges of each histogram bar. Ideal for unequal spacings between histograms.
    ax : class
        The matplotlib plotting axis.
    color : str
        The color of the histogram error plot.
    alpha : float
        The transparency fraction of the histogram error plot.
    """
    plot_histogram_confidence(x_mid, histogram-histogram_error, histogram+histogram_error,
                              x_edges=x_edges, ax=ax, color=color, alpha=alpha)


class PlotHistMST:

    def _get_rotate_colors(self):
        """Cycles through the default matplotlib colours, i.e. C0 - C9.
        """
        color = 'C'+str(self.rotate_colors)
        if self.rotate_colors + 1 == 10:
            self.rotate_colors = 0
        else:
            self.rotate_colors += 1
        return color

    def _get_rotate_linestyles(self):
        """Cycles through linestyles "-", "--", "-." and ":".
        """
        linestyle = self.linestyle_all[self.rotate_linestyle]
        if self.rotate_linestyle + 1 == 4:
            self.rotate_linestyle = 0
        else:
            self.rotate_linestyle += 1
        return linestyle

    def __init__(self):
        """Initialises the plotting class.
        """
        # tracking data
        self.num_data = 0
        # line information
        self.colors = []
        self.alphas = []
        self.alphas_envelope = []
        self.labels = []
        self.linestyles = []
        self.linewidths = []
        self.binned_data = []
        self.need_envelopes = []
        # colors and styles
        self.rotate_colors = 0
        self.rotate_linestyle = 0
        self.linestyle_all = ['-', '--', '-.', ':']
        # Type of data used
        self.use_sqrt_s = None
        self.uselog = None
        # Binning information
        self.d_min = None
        self.d_max = None
        self.l_min = None
        self.l_max = None
        self.b_min = None
        self.b_max = None
        self.s_min = None
        self.s_max = None
        self.l_edges = None
        self.b_edges = None
        self.s_edges = None
        self.usenorm = None

    def read_mst(self, mst_hist, color=None, linewidth=2., linestyle=None, alpha=0.8,
                 label=None, alpha_envelope=0.3):
        """Input minimum spanning tree statistics.

        Parameters
        ----------
        mst_hist : dict
            Binned MST dictionary, given in the format outputted by HistMST.
        color : str
            Color of the MST histogram.
        linewidth : float
            Width of line used.
        linestyle : str
            Linestyle used.
        alpha : float
            The transparency of the line.
        label : str
            label used in the legend, this will only appear in the legend if this is given.
        alpha_envelope : float
            The transparency of envelopes (usually the standard deviation of the counts in each bin).
        """
        if self.usenorm is None:
            self.usenorm = mst_hist['usenorm']
        self.binned_data.append(mst_hist)
        self.num_data += 1
        delta_d = mst_hist['x_d'][1]-mst_hist['x_d'][0]
        self.d_min = mst_hist['x_d'][0] - delta_d/2.
        self.d_max = mst_hist['x_d'][-1] + delta_d/2.
        if mst_hist['uselog'] == False:
            self.uselog = False
            delta_l = mst_hist['x_l'][1]-mst_hist['x_l'][0]
            self.l_min = mst_hist['x_l'][0] - delta_l/2.
            self.l_max = mst_hist['x_l'][-1] + delta_l/2.
            delta_b = mst_hist['x_b'][1]-mst_hist['x_l'][0]
            self.b_min = mst_hist['x_b'][0] - delta_b/2.
            self.b_max = mst_hist['x_b'][-1] + delta_b/2.
            self.l_edges = np.linspace(self.l_min, self.l_max, len(mst_hist['x_l'])+1)
            self.b_edges = np.linspace(self.b_min, self.b_max, len(mst_hist['x_b'])+1)
        else:
            self.uselog = True
            logdelta_l = np.log10(mst_hist['x_l'][1])-np.log10(mst_hist['x_l'][0])
            self.l_min = 10.**(np.log10(mst_hist['x_l'][0]) - logdelta_l/2.)
            self.l_max = 10.**(np.log10(mst_hist['x_l'][-1]) + logdelta_l/2.)
            logdelta_b = np.log10(mst_hist['x_b'][1])-np.log10(mst_hist['x_b'][0])
            self.b_min = 10.**(np.log10(mst_hist['x_b'][0]) - logdelta_b/2.)
            self.b_max = 10.**(np.log10(mst_hist['x_b'][-1]) + logdelta_b/2.)
            logl_edges = np.linspace(np.log10(self.l_min), np.log10(self.l_max), len(mst_hist['x_l'])+1)
            logb_edges = np.linspace(np.log10(self.b_min), np.log10(self.b_max), len(mst_hist['x_b'])+1)
            self.l_edges = 10.**logl_edges
            self.b_edges = 10.**logb_edges
        delta_s = mst_hist['x_s'][1]-mst_hist['x_s'][0]
        self.s_min = 0.
        self.s_max = 1.
        if mst_hist['use_sqrt_s'] == True:
            self.use_sqrt_s = True
        if color is None:
            self.colors.append(self._get_rotate_colors())
        else:
            self.colors.append(color)
        self.alphas.append(alpha)
        self.alphas_envelope.append(alpha_envelope)
        self.linewidths.append(linewidth)
        if linestyle is None:
            self.linestyles.append(self._get_rotate_linestyles())
        else:
            self.linestyles.append(linestyle)
        self.labels.append(label)
        self.need_envelopes.append(mst_hist['isgroup'])

    def plot(self, usebox=True, saveas=None, fontsize=16, figsize=(16, 4), units=None,
             showenvelopes=True, usecomp=False, usemean=True, height_ratios=[2, 1],
             usefraction=False, whichcomp=0, plotzeroline=True, legend=True, subplot_adjust_top=0.85,
             legend_fontsize=14, legend_column=4, xlabels=[None, None, None, None], plt_output='show'):
        """Outputs the final plot of the MST statistics.

        Parameters
        ----------
        usebox : bool
            For l, b and s this sets whether to use boxes for the histogram or to simple plot as line
            from the bin centre.
        saveas : str
            If not None, this will save the plot with the name provided.
        fontsize : int
            Fontsize of axis labels.
        figsize : tuple
            Dimensions of the figure.
        units : str
            Units of l and b MST statistics, if None is supplied then we assume it is unitless.
        showenvelopes : bool
            This determines whether to plot data with input standard deviation.
        usecomp : bool
            Determines whether to include comparison subplots.
        usemean : bool
            For comparison plots, the determines whether to use the mean of input distributions.
        height_ratios : tuple
            Height ratio between main plots and comparison plots.
        usefraction : bool
            If true comparison plots are given by the fractional difference to a reference data otherwise
            they are given as absolute differences.
        whichcomp : int
            Determines which data should be the reference data that all other data is compared to. Only used
            if usemean == False.
        plotzeroline : bool
            Plots horizontal zero line for subplot.
        legend : bool
            Determines whether to include a legend.
        subplot_adjust_top : float
            If legend == True then the figure's top is adjusted by the input value.
        legend_fontsize : int
            Size of legend text.
        legend_column : int
            Number of keys in each line of the legend.
        xlabels : list
            List of string labels to replace the default if not set to None.
        """
        # setting up figure
        if self.usenorm == True:
            ylabel = '\\bar{N}'
        else:
            ylabel = 'N'
        if usecomp == True and self.num_data > 1:
            f, ((ax1, ax2, ax3, ax4),
                (ax15, ax25, ax35, ax45)) = plt.subplots(2, 4, figsize=figsize,
                                                         gridspec_kw={'height_ratios': height_ratios})
            if xlabels[0] is None:
                ax15.set_xlabel(r'$d$', fontsize=fontsize)
            else:
                ax15.set_xlabel(xlabels[0], fontsize=fontsize)
            if units is None:
                if xlabels[1] is None:
                    ax25.set_xlabel(r'$l$', fontsize=fontsize)
                else:
                    ax25.set_xlabel(xlabels[1], fontsize=fontsize)
                if xlabels[2] is None:
                    ax35.set_xlabel(r'$b$', fontsize=fontsize)
                else:
                    ax35.set_xlabel(xlabels[2], fontsize=fontsize)
            else:
                if xlabels[1] is None:
                    ax25.set_xlabel(r'$l$ $[%s]$' %(units), fontsize=fontsize)
                else:
                    ax25.set_xlabel(xlabels[1]+r' $[%s]$' %(units), fontsize=fontsize)
                if xlabels[2] is None:
                    ax35.set_xlabel(r'$b$ $[%s]$' %(units), fontsize=fontsize)
                else:
                    ax35.set_xlabel(xlabels[2]+r' $[%s]$' %(units), fontsize=fontsize)
            if self.use_sqrt_s == True:
                if xlabels[3] is None:
                    ax45.set_xlabel(r'$\sqrt{1-s}$', fontsize=fontsize)
                else:
                    ax45.set_xlabel(xlabels[3], fontsize=fontsize)
            else:
                if xlabels[3] is None:
                    ax45.set_xlabel(r'$s$', fontsize=fontsize)
                else:
                    ax45.set_xlabel(xlabels[3], fontsize=fontsize)
            if usefraction == False:
                ax15.set_ylabel(r'$\Delta %s$' %(ylabel), fontsize=fontsize)
                ax25.set_ylabel(r'$\Delta %s$' %(ylabel), fontsize=fontsize)
                ax35.set_ylabel(r'$\Delta %s$' %(ylabel), fontsize=fontsize)
                ax45.set_ylabel(r'$\Delta %s$' %(ylabel), fontsize=fontsize)
            else:
                if usemean == False:
                    if self.labels[0] is None:
                        comp_label = str(0)
                    else:
                        comp_label = str(self.labels[0])
                else:
                    comp_label = '{\\rm Mean}'
                ax15.set_ylabel(r'$\frac{%s}{%s_{%s}}-1$' %(ylabel, ylabel, comp_label), fontsize=fontsize)
                ax25.set_ylabel(r'$\frac{%s}{%s_{%s}}-1$' %(ylabel, ylabel, comp_label), fontsize=fontsize)
                ax35.set_ylabel(r'$\frac{%s}{%s_{%s}}-1$' %(ylabel, ylabel, comp_label), fontsize=fontsize)
                ax45.set_ylabel(r'$\frac{%s}{%s_{%s}}-1$' %(ylabel, ylabel, comp_label), fontsize=fontsize)
        else:
            f, (ax1, ax2, ax3, ax4) = plt.subplots(1, 4, figsize=figsize)
            ax1.set_xlabel(r'$d$', fontsize=fontsize)
            if units is None:
                ax2.set_xlabel(r'$l$', fontsize=fontsize)
                ax3.set_xlabel(r'$b$', fontsize=fontsize)
            else:
                ax2.set_xlabel(r'$l$ $[%s]$' %(units), fontsize=fontsize)
                ax3.set_xlabel(r'$b$ $[%s]$' %(units), fontsize=fontsize)
            if self.use_sqrt_s == True:
                ax4.set_xlabel(r'$\sqrt{1-s}$', fontsize=fontsize)
            else:
                ax4.set_xlabel(r'$s$', fontsize=fontsize)
        ax1.set_ylabel(r'$%s$' %(ylabel), fontsize=fontsize)
        ax2.set_ylabel(r'$%s$' %(ylabel), fontsize=fontsize)
        ax3.set_ylabel(r'$%s$' %(ylabel), fontsize=fontsize)
        ax4.set_ylabel(r'$%s$' %(ylabel), fontsize=fontsize)
        if self.uselog == True:
            ax2.set_xscale('log')
            ax3.set_xscale('log')
            if usecomp == True and self.num_data > 1:
                ax25.set_xscale('log')
                ax35.set_xscale('log')
        # Main Plots
        if usecomp == True and self.num_data > 1:
            if usemean == False:
                y_d_comp = self.binned_data[whichcomp]['y_d']
                y_l_comp = self.binned_data[whichcomp]['y_l']
                y_b_comp = self.binned_data[whichcomp]['y_b']
                y_s_comp = self.binned_data[whichcomp]['y_s']
            else:
                y_d_all = []
                y_l_all = []
                y_b_all = []
                y_s_all = []
                for i in range(0, len(self.binned_data)):
                    y_d_all.append(self.binned_data[i]['y_d'])
                    y_l_all.append(self.binned_data[i]['y_l'])
                    y_b_all.append(self.binned_data[i]['y_b'])
                    y_s_all.append(self.binned_data[i]['y_s'])
                y_d_all = np.array(y_d_all)
                y_l_all = np.array(y_l_all)
                y_b_all = np.array(y_b_all)
                y_s_all = np.array(y_s_all)
                y_d_comp = np.mean(y_d_all, axis=0)
                y_l_comp = np.mean(y_l_all, axis=0)
                y_b_comp = np.mean(y_b_all, axis=0)
                y_s_comp = np.mean(y_s_all, axis=0)
            if usefraction == True:
                c1 = np.where(y_d_comp != 0.)[0]
                c2 = np.where(y_l_comp != 0.)[0]
                c3 = np.where(y_b_comp != 0.)[0]
                c4 = np.where(y_s_comp != 0.)[0]
        for i in range(0, len(self.binned_data)):
            plot_histogram_line(self.binned_data[i]['x_d'], self.binned_data[i]['y_d'],
                                ax=ax1, color=self.colors[i], alpha=self.alphas[i],
                                linewidth=self.linewidths[i], linestyle=self.linestyles[i])
            if showenvelopes == True and self.need_envelopes[i] == True:
                plot_histogram_error(self.binned_data[i]['x_d'], self.binned_data[i]['y_d'], self.binned_data[i]['y_d_std'],
                                     ax=ax1, color=self.colors[i], alpha=self.alphas_envelope[i])
            if usebox == True:
                plot_histogram_line(self.binned_data[i]['x_l'], self.binned_data[i]['y_l'], x_edges=self.l_edges,
                                    ax=ax2, color=self.colors[i], alpha=self.alphas[i],
                                    linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                plot_histogram_line(self.binned_data[i]['x_b'], self.binned_data[i]['y_b'], x_edges=self.b_edges,
                                    ax=ax3, color=self.colors[i], alpha=self.alphas[i],
                                    linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                plot_histogram_line(self.binned_data[i]['x_s'], self.binned_data[i]['y_s'],
                                    ax=ax4, color=self.colors[i], alpha=self.alphas[i],
                                    linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                if showenvelopes == True and self.need_envelopes[i] == True:
                    plot_histogram_error(self.binned_data[i]['x_l'], self.binned_data[i]['y_l'], self.binned_data[i]['y_l_std'],
                                         x_edges=self.l_edges, ax=ax2, color=self.colors[i], alpha=self.alphas_envelope[i])
                    plot_histogram_error(self.binned_data[i]['x_b'], self.binned_data[i]['y_b'], self.binned_data[i]['y_b_std'],
                                         x_edges=self.b_edges, ax=ax3, color=self.colors[i], alpha=self.alphas_envelope[i])
                    plot_histogram_error(self.binned_data[i]['x_s'], self.binned_data[i]['y_s'], self.binned_data[i]['y_s_std'],
                                         ax=ax4, color=self.colors[i], alpha=self.alphas_envelope[i])
            else:
                ax2.plot(self.binned_data[i]['x_l'], self.binned_data[i]['y_l'],
                         color=self.colors[i], alpha=self.alphas[i],
                         linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                ax3.plot(self.binned_data[i]['x_b'], self.binned_data[i]['y_b'],
                         color=self.colors[i], alpha=self.alphas[i],
                         linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                ax4.plot(self.binned_data[i]['x_s'], self.binned_data[i]['y_s'],
                         color=self.colors[i], alpha=self.alphas[i],
                         linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                if showenvelopes == True and self.need_envelopes[i] == True:
                    ax2.fill_between(self.binned_data[i]['x_l'], self.binned_data[i]['y_l']-self.binned_data[i]['y_l_std'],
                                     self.binned_data[i]['y_l']+self.binned_data[i]['y_l_std'], color=self.colors[i],
                                     alpha=self.alphas_envelope[i])
                    ax3.fill_between(self.binned_data[i]['x_b'], self.binned_data[i]['y_b']-self.binned_data[i]['y_b_std'],
                                     self.binned_data[i]['y_b']+self.binned_data[i]['y_b_std'], color=self.colors[i],
                                     alpha=self.alphas_envelope[i])
                    ax4.fill_between(self.binned_data[i]['x_s'], self.binned_data[i]['y_s']-self.binned_data[i]['y_s_std'],
                                     self.binned_data[i]['y_s']+self.binned_data[i]['y_s_std'], color=self.colors[i],
                                     alpha=self.alphas_envelope[i])
            if usecomp == True and self.num_data > 1:
                if usefraction == False:
                    plot_histogram_line(self.binned_data[i]['x_d'], self.binned_data[i]['y_d']-y_d_comp,
                                        ax=ax15, color=self.colors[i], alpha=self.alphas[i],
                                        linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                    if showenvelopes == True and self.need_envelopes[i] == True:
                        plot_histogram_error(self.binned_data[i]['x_d'], self.binned_data[i]['y_d']-y_d_comp, self.binned_data[i]['y_d_std'],
                                             ax=ax15, color=self.colors[i], alpha=self.alphas_envelope[i])
                    if usebox == True:
                        plot_histogram_line(self.binned_data[i]['x_l'], self.binned_data[i]['y_l']-y_l_comp,
                                            x_edges=self.l_edges, ax=ax25, color=self.colors[i], alpha=self.alphas[i],
                                            linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        plot_histogram_line(self.binned_data[i]['x_b'], self.binned_data[i]['y_b']-y_b_comp,
                                            x_edges=self.b_edges, ax=ax35, color=self.colors[i], alpha=self.alphas[i],
                                            linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        plot_histogram_line(self.binned_data[i]['x_s'], self.binned_data[i]['y_s']-y_s_comp,
                                            ax=ax45, color=self.colors[i], alpha=self.alphas[i],
                                            linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        if showenvelopes == True and self.need_envelopes[i] == True:
                            plot_histogram_error(self.binned_data[i]['x_l'], self.binned_data[i]['y_l']-y_l_comp,
                                                 self.binned_data[i]['y_l_std'], x_edges=self.l_edges,
                                                 ax=ax25, color=self.colors[i], alpha=self.alphas_envelope[i])
                            plot_histogram_error(self.binned_data[i]['x_b'], self.binned_data[i]['y_b']-y_b_comp,
                                                 self.binned_data[i]['y_b_std'], x_edges=self.b_edges,
                                                 ax=ax35, color=self.colors[i], alpha=self.alphas_envelope[i])
                            plot_histogram_error(self.binned_data[i]['x_s'], self.binned_data[i]['y_s']-y_s_comp,
                                                 self.binned_data[i]['y_s_std'],
                                                 ax=ax45, color=self.colors[i], alpha=self.alphas_envelope[i])
                    else:
                        ax25.plot(self.binned_data[i]['x_l'], self.binned_data[i]['y_l']-y_l_comp,
                                  color=self.colors[i], alpha=self.alphas[i],
                                  linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        ax35.plot(self.binned_data[i]['x_b'], self.binned_data[i]['y_b']-y_b_comp,
                                  color=self.colors[i], alpha=self.alphas[i],
                                  linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        ax45.plot(self.binned_data[i]['x_s'], self.binned_data[i]['y_s']-y_s_comp,
                                  color=self.colors[i], alpha=self.alphas[i],
                                  linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        if showenvelopes == True and self.need_envelopes[i] == True:
                            ax25.fill_between(self.binned_data[i]['x_l'], self.binned_data[i]['y_l']-self.binned_data[i]['y_l_std']-y_l_comp,
                                              self.binned_data[i]['y_l']+self.binned_data[i]['y_l_std']-y_l_comp,
                                              color=self.colors[i], alpha=self.alphas_envelope[i])
                            ax35.fill_between(self.binned_data[i]['x_b'], self.binned_data[i]['y_b']-self.binned_data[i]['y_b_std']-y_b_comp,
                                              self.binned_data[i]['y_b']+self.binned_data[i]['y_b_std']-y_b_comp,
                                              color=self.colors[i], alpha=self.alphas_envelope[i])
                            ax45.fill_between(self.binned_data[i]['x_s'], self.binned_data[i]['y_s']-self.binned_data[i]['y_s_std']-y_s_comp,
                                              self.binned_data[i]['y_s']+self.binned_data[i]['y_s_std']-y_s_comp, color=self.colors[i],
                                              alpha=self.alphas_envelope[i])
                else:
                    plot_histogram_line(self.binned_data[i]['x_d'][c1], self.binned_data[i]['y_d'][c1]/y_d_comp[c1]-1.,
                                        ax=ax15, color=self.colors[i], alpha=self.alphas[i],
                                        linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                    if showenvelopes == True and self.need_envelopes[i] == True:
                        plot_histogram_confidence(self.binned_data[i]['x_d'][c1],
                                                  (self.binned_data[i]['y_d'][c1]-self.binned_data[i]['y_d_std'][c1])/y_d_comp[c1]-1,
                                                  (self.binned_data[i]['y_d'][c1]+self.binned_data[i]['y_d_std'][c1])/y_d_comp[c1]-1,
                                                  ax=ax15, color=self.colors[i], alpha=self.alphas_envelope[i])
                    if usebox == True:
                        plot_histogram_line(self.binned_data[i]['x_l'][c2], self.binned_data[i]['y_l'][c2]/y_l_comp[c2]-1.,
                                            x_edges=self.l_edges, ax=ax25, color=self.colors[i], alpha=self.alphas[i],
                                            linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        plot_histogram_line(self.binned_data[i]['x_b'][c3], self.binned_data[i]['y_b'][c3]/y_b_comp[c3]-1.,
                                            x_edges=self.b_edges, ax=ax35, color=self.colors[i], alpha=self.alphas[i],
                                            linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        plot_histogram_line(self.binned_data[i]['x_s'][c4], self.binned_data[i]['y_s'][c4]/y_s_comp[c4]-1,
                                            ax=ax45, color=self.colors[i], alpha=self.alphas[i],
                                            linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        if showenvelopes == True and self.need_envelopes[i] == True:
                            plot_histogram_confidence(self.binned_data[i]['x_l'][c2],
                                                      (self.binned_data[i]['y_l'][c2]-self.binned_data[i]['y_l_std'][c2])/y_l_comp[c2]-1,
                                                      (self.binned_data[i]['y_l'][c2]+self.binned_data[i]['y_l_std'][c2])/y_l_comp[c2]-1,
                                                      x_edges=self.l_edges, ax=ax25, color=self.colors[i], alpha=self.alphas_envelope[i])
                            plot_histogram_confidence(self.binned_data[i]['x_b'][c3],
                                                      (self.binned_data[i]['y_b'][c3]-self.binned_data[i]['y_b_std'][c3])/y_b_comp[c3]-1,
                                                      (self.binned_data[i]['y_b'][c3]+self.binned_data[i]['y_b_std'][c3])/y_b_comp[c3]-1,
                                                      x_edges=self.b_edges, ax=ax35, color=self.colors[i], alpha=self.alphas_envelope[i])
                            plot_histogram_confidence(self.binned_data[i]['x_s'][c4],
                                                      (self.binned_data[i]['y_s'][c4]-self.binned_data[i]['y_s_std'][c4])/y_s_comp[c4]-1,
                                                      (self.binned_data[i]['y_s'][c4]+self.binned_data[i]['y_s_std'][c4])/y_s_comp[c4]-1,
                                                      ax=ax45, color=self.colors[i], alpha=self.alphas_envelope[i])
                    else:
                        ax25.plot(self.binned_data[i]['x_l'][c2], self.binned_data[i]['y_l'][c2]/y_l_comp[c2]-1.,
                                 color=self.colors[i], alpha=self.alphas[i],
                                 linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        ax35.plot(self.binned_data[i]['x_b'][c3], self.binned_data[i]['y_b'][c3]/y_b_comp[c3]-1.,
                                 color=self.colors[i], alpha=self.alphas[i],
                                 linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        ax45.plot(self.binned_data[i]['x_s'][c4], self.binned_data[i]['y_s'][c4]/y_s_comp[c4]-1.,
                                 color=self.colors[i], alpha=self.alphas[i],
                                 linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        if showenvelopes == True and self.need_envelopes[i] == True:
                            ax25.fill_between(self.binned_data[i]['x_l'][c2],
                                              (self.binned_data[i]['y_l'][c2]-self.binned_data[i]['y_l_std'][c2])/y_l_comp[c2]-1.,
                                              (self.binned_data[i]['y_l'][c2]+self.binned_data[i]['y_l_std'][c2])/y_l_comp[c2]-1.,
                                              color=self.colors[i], alpha=self.alphas_envelope[i],
                                              linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                            ax35.fill_between(self.binned_data[i]['x_b'][c3],
                                              (self.binned_data[i]['y_b'][c3]-self.binned_data[i]['y_b_std'][c3])/y_b_comp[c3]-1.,
                                              (self.binned_data[i]['y_b'][c3]+self.binned_data[i]['y_b_std'][c3])/y_b_comp[c3]-1.,
                                              color=self.colors[i], alpha=self.alphas_envelope[i],
                                              linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                            ax45.fill_between(self.binned_data[i]['x_s'][c4],
                                              (self.binned_data[i]['y_s'][c4]-self.binned_data[i]['y_s_std'][c4])/y_s_comp[c4]-1.,
                                              (self.binned_data[i]['y_s'][c4]+self.binned_data[i]['y_s_std'][c4])/y_s_comp[c4]-1.,
                                              color=self.colors[i], alpha=self.alphas_envelope[i],
                                              linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                if plotzeroline == True:
                    ax15.axhline(0., color='k', linestyle=':', linewidth=1.)
                    ax25.axhline(0., color='k', linestyle=':', linewidth=1.)
                    ax35.axhline(0., color='k', linestyle=':', linewidth=1.)
                    ax45.axhline(0., color='k', linestyle=':', linewidth=1.)
        ax1.set_xlim(self.d_min, self.d_max)
        ax2.set_xlim(self.l_min, self.l_max)
        ax3.set_xlim(self.b_min, self.b_max)
        ax4.set_xlim(self.s_min, self.s_max)
        ax1.set_ylim(0., None)
        ax2.set_ylim(0., None)
        ax3.set_ylim(0., None)
        ax4.set_ylim(0., None)
        if usecomp == True and self.num_data > 1:
            ax1.set_xticks([])
            ax2.set_xticks([])
            ax3.set_xticks([])
            ax4.set_xticks([])
            ax15.set_xlim(self.d_min, self.d_max)
            ax25.set_xlim(self.l_min, self.l_max)
            ax35.set_xlim(self.b_min, self.b_max)
            ax45.set_xlim(self.s_min, self.s_max)
        plt.tight_layout()
        if legend == True:
            lines = []
            labels = []
            for i in range(0, len(self.binned_data)):
                if self.labels[i] is None:
                    pass
                else:
                    if showenvelopes == True and self.need_envelopes[i] == True:
                        line = ax1.fill_between([], [], [], color=self.colors[i], alpha=self.alphas_envelope[i])
                        lines.append(line)
                        labels.append(self.labels[i])
                    else:
                        line, = ax1.plot([], [], color=self.colors[i], alpha=self.alphas[i],
                                         linewidth=self.linewidths[i], linestyle=self.linestyles[i])
                        lines.append(line)
                        labels.append(self.labels[i])
            f.subplots_adjust(top=subplot_adjust_top)
            plt.figlegend(handles=lines, labels=labels, loc='upper center', fontsize=legend_fontsize, ncol=legend_column, frameon=False)
        if saveas is None:
            pass
        else:
            plt.savefig(saveas)
        if plt_output == 'show':
            plt.show()
        elif plt_output == 'close':
            plt.close()
        else:
            pass

    def clean(self):
        """Resets PlotHistMST variables.
        """
        self.__init__()



#'graph.py' contains functions for switching how the edges of a graph are stored,
# i.e. either as a csr_matrix or a 2 dimensional array.

from scipy.sparse import csr_matrix


def graph2data(graph):
    """Returns the index and data of a sparse matrix.

    Parameters
    ----------
    graph : csr_matrix
        A sparse matrix of the edges in a graph and corresponding node indexes.

    Returns
    -------
    index1, index2 : array
        Arrays of the node indices of the graph.
    weights : array
        An array of the weights of each edge in the graph.
    """
    graph = graph.tocoo()
    weights = graph.data
    index1 = graph.row
    index2 = graph.col
    return index1, index2, weights


def data2graph(index1, index2, weights, num_nodes):
    """Returns the sparse matrix of a graph given the indices and data.

    Parameters
    ----------
    index1, index2 : array
        Arrays of the node indices of the graph.
    weights : array
        An array of the weights of each edge in the graph.
    num_nodes : int
        Number of nodes.

    Returns
    -------
    graph : csr_matrix
        A sparse matrix of the edges in a graph and corresponding node indexes.
    """
    graph = csr_matrix((weights, (index1, index2)), shape=(num_nodes, num_nodes))
    return graph
